# -*- coding: utf-8 -*-
import json
import os

import mock

from ddtrace.appsec._remoteconfiguration import RCASMDDCallBack
from ddtrace.internal import runtime
from ddtrace.internal.remoteconfig.client import RemoteConfigClient
from ddtrace.internal.utils.version import _pep440_to_semver
from tests.utils import override_env


def _expected_payload(
    rc_client,
    has_errors=False,
    targets_version=0,
    backend_client_state=None,
    config_states=[],
    cached_target_files=[],
    error_msg=None,
):
    payload = {
        "client": {
            "id": rc_client.id,
            "products": ["ASM_FEATURES"],
            "is_tracer": True,
            "client_tracer": {
                "runtime_id": runtime.get_runtime_id(),
                "language": "python",
                "tracer_version": _pep440_to_semver(),
                "service": None,
                "env": None,
                "app_version": None,
            },
            "state": {
                "root_version": 1,
                "targets_version": targets_version,
                "config_states": config_states,
                "has_error": has_errors,
            },
            "capabilities": "Ag==",
        },
        "cached_target_files": cached_target_files,
    }
    if backend_client_state:
        payload["client"]["state"]["backend_client_state"] = backend_client_state
    if has_errors:
        payload["client"]["state"]["error"] = error_msg
    return payload


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
MOCK_AGENT_RESPONSES_FILE = os.path.join(ROOT_DIR, "rc_mocked_responses_asm_features.json")


def _assert_response(mock_send_request, expected_response):
    expected_response["cached_target_files"].sort(key=lambda x: x["path"], reverse=True)
    expected_response["client"]["state"]["config_states"].sort(key=lambda x: x["id"], reverse=True)
    response = json.loads(mock_send_request.call_args.args[0])
    response["cached_target_files"].sort(key=lambda x: x["path"], reverse=True)
    response["client"]["state"]["config_states"].sort(key=lambda x: x["id"], reverse=True)

    assert response["client"]["client_tracer"]["tags"]
    del response["client"]["client_tracer"]["tags"]

    assert response == expected_response


@mock.patch.object(RemoteConfigClient, "_send_request")
@mock.patch("ddtrace.internal.remoteconfig.client._appsec_rc_capabilities")
def test_remote_config_client_steps(mock_appsec_rc_capabilities, mock_send_request):
    with open(MOCK_AGENT_RESPONSES_FILE, "r") as f:
        MOCK_AGENT_RESPONSES = json.load(f)

    class RCAppSecMockCallBack(RCASMDDCallBack):
        def __call__(self, metadata, features):
            mock_callback(metadata, features)

    mock_appsec_rc_capabilities.return_value = "Ag=="

    rc_client = RemoteConfigClient()
    mock_callback = mock.mock.MagicMock()
    callback = RCAppSecMockCallBack(mock.mock.MagicMock())

    rc_client.register_product("ASM_FEATURES", callback)
    with override_env(dict(DD_REMOTE_CONFIGURATION_ENABLED="false")):
        # 0.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[0]
        rc_client.request()
        expected_response = _expected_payload(rc_client)

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_not_called()
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 1. An update that doesn’t have any new config files but does have an updated TUF Targets file.
        # The tracer is supposed to process this update and store that the latest TUF Targets version is 1.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[1]
        rc_client.request()
        expected_response = _expected_payload(rc_client, targets_version=1, backend_client_state="eyJmb28iOiAiYmFyIn0=")

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_called_with(mock.ANY, {"asm": {"enabled": True}})
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 2. A single configuration for the product is added. (“base”)
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[2]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=2,
            config_states=[{"id": "ASM_FEATURES-base", "version": 1, "product": "ASM_FEATURES"}],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                }
            ],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_called_with(mock.ANY, {"asm": {"enabled": False}})
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 3. The “base” configuration is modified.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[3]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=3,
            config_states=[{"id": "ASM_FEATURES-base", "version": 2, "product": "ASM_FEATURES"}],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 48,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "a38ebf9fa256071f9823a5f512a84e8a786c8da2b0719452deeb5dc287ec990f",
                        }
                    ],
                }
            ],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_called_with(mock.ANY, False)
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 4. The “base” configuration is removed.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[4]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=4,
            config_states=[],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_called_with(mock.ANY, {"asm": {"enabled": True}})
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 5. The “base” configuration is added along with the “second” configuration.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[5]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=5,
            config_states=[
                {"id": "ASM_FEATURES-base", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-second", "version": 1, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-second/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_called_with(mock.ANY, {"asm": {"enabled": True}})
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 6. The “third” configuration is added, the “second” configuration is removed
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[6]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=6,
            config_states=[
                {"id": "ASM_FEATURES-base", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-third", "version": 1, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-third/config",
                    "length": 41,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "1e4a75edfd5f65e0adec704cc7c32f79987117d84755544ae4905045cdb0a443",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_called()
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 7. The “second” configuration is added back. The “first” configuration is modified.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[7]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=7,
            config_states=[
                {"id": "ASM_FEATURES-third", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-base", "version": 2, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-second", "version": 1, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-third/config",
                    "length": 41,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "1e4a75edfd5f65e0adec704cc7c32f79987117d84755544ae4905045cdb0a443",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 48,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "a38ebf9fa256071f9823a5f512a84e8a786c8da2b0719452deeb5dc287ec990f",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-second/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_called_with(mock.ANY, {"asm": {"enabled": True}})
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 8. The “first” configuration is modified again, the “third” configuration is removed.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[8]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=8,
            config_states=[
                {"id": "ASM_FEATURES-second", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-base", "version": 1, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-second/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_not_called()
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 9. Another update that doesn’t have any new or updated config files but has a newer TUF Targets file.
        # This tests that a tracer handles this scenario and still reports that it has tracked config files in
        # the next update.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[9]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=9,
            config_states=[
                {"id": "ASM_FEATURES-second", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-base", "version": 1, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-second/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_not_called()
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 10. A file is included in TUF Targets that is NOT specified in client_configs.
        # The tracer should ignore this file and not report it in the next update request.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[10]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=10,
            config_states=[
                {"id": "ASM_FEATURES-second", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-base", "version": 1, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-second/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_called_with(mock.ANY, {"asm": {"enabled": True}})
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 11. The “first” configuration is modified. Another file is included in the TUF Targets that is not specified
        # in client_configs. This again tests that the tracer is only processing files in client_configs.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[11]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=11,
            config_states=[
                {"id": "ASM_FEATURES-second", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-base", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-third", "version": 1, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-second/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-third/testname",
                    "length": 41,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "1e4a75edfd5f65e0adec704cc7c32f79987117d84755544ae4905045cdb0a443",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error is None
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_called_with(mock.ANY, {"asm": {"enabled": False}})
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 12. A new configuration file’s raw bytes are missing from target_files that is referenced by client_configs
        # and targets.signed.targets. This update should fail. The tracer client’s state should not change other
        # than reporting the error in the has_error field along with a message in the error field of the next request.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[12]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=12,
            config_states=[
                {"id": "ASM_FEATURES-second", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-third", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-base", "version": 2, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-second/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-third/testname",
                    "length": 41,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "1e4a75edfd5f65e0adec704cc7c32f79987117d84755544ae4905045cdb0a443",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 48,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "a38ebf9fa256071f9823a5f512a84e8a786c8da2b0719452deeb5dc287ec990f",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error == "Not all client configurations have target files"
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_not_called()
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 13. A new configuration file is missing the TUF metadata in targets.signed.targets but is referenced in
        # client_configs and target_files. This update should fail. The tracer client’s state should not change
        # other than reporting the error in the has_error field along with a message in the error field of the next
        # request.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[13]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=12,
            has_errors=True,
            error_msg="Not all client configurations have target files",
            config_states=[
                {"id": "ASM_FEATURES-second", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-third", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-base", "version": 2, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-second/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-third/testname",
                    "length": 41,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "1e4a75edfd5f65e0adec704cc7c32f79987117d84755544ae4905045cdb0a443",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 48,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "a38ebf9fa256071f9823a5f512a84e8a786c8da2b0719452deeb5dc287ec990f",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error == (
            "target file datadog/2/ASM_FEATURES/ASM_FEATURES-third/testname "
            "not exists in client_config and signed targets"
        )
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_not_called()
        mock_send_request.reset_mock()
        mock_callback.reset_mock()

        # 14.
        mock_send_request.return_value = MOCK_AGENT_RESPONSES[13]
        rc_client.request()
        expected_response = _expected_payload(
            rc_client,
            targets_version=12,
            has_errors=True,
            error_msg=(
                "target file datadog/2/ASM_FEATURES/ASM_FEATURES-third/testname "
                "not exists in client_config and signed targets"
            ),
            config_states=[
                {"id": "ASM_FEATURES-second", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-third", "version": 1, "product": "ASM_FEATURES"},
                {"id": "ASM_FEATURES-base", "version": 2, "product": "ASM_FEATURES"},
            ],
            backend_client_state="eyJmb28iOiAiYmFyIn0=",
            cached_target_files=[
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-second/config",
                    "length": 47,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "9221dfd9f6084151313e3e4920121ae843614c328e4630ea371ba66e2f15a0a6",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-third/testname",
                    "length": 41,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "1e4a75edfd5f65e0adec704cc7c32f79987117d84755544ae4905045cdb0a443",
                        }
                    ],
                },
                {
                    "path": "datadog/2/ASM_FEATURES/ASM_FEATURES-base/config",
                    "length": 48,
                    "hashes": [
                        {
                            "algorithm": "sha256",
                            "hash": "a38ebf9fa256071f9823a5f512a84e8a786c8da2b0719452deeb5dc287ec990f",
                        }
                    ],
                },
            ],
        )

        assert rc_client._last_error == (
            "target file datadog/2/ASM_FEATURES/ASM_FEATURES-third/testname "
            "not exists in client_config and signed targets"
        )
        _assert_response(mock_send_request, expected_response)
        mock_callback.assert_not_called()
        mock_send_request.reset_mock()
        mock_callback.reset_mock()



import inspect
import sys

from ddtrace.internal.wrapping import unwrap
from ddtrace.internal.wrapping import wrap


def test_wrap_unwrap():
    channel = []

    def wrapper(f, args, kwargs):
        channel.append((args, kwargs))
        retval = f(*args, **kwargs)
        channel.append(retval)
        return retval

    def f(a, b, c=None):
        return (a, b, c)

    wrap(f, wrapper)

    assert f(1, 2, 3) == (1, 2, 3)
    assert channel == [((1, 2, 3), {}), (1, 2, 3)]

    assert f(1, b=2, c=3) == (1, 2, 3)
    assert channel == [((1, 2, 3), {}), (1, 2, 3), ((1, 2, 3), {}), (1, 2, 3)]

    channel[:] = []

    unwrap(f, wrapper)

    assert f(1, 2, 3) == (1, 2, 3)
    assert not channel


def test_mutiple_wrap():
    channel1, channel2 = [], []

    def wrapper_maker(channel):
        def wrapper(f, args, kwargs):
            channel[:] = []
            channel.append((args, kwargs))
            retval = f(*args, **kwargs)
            channel.append(retval)
            return retval

        return wrapper

    wrapper1, wrapper2 = (wrapper_maker(_) for _ in (channel1, channel2))

    def f(a, b, c=None):
        return (a, b, c)

    wrap(f, wrapper1)
    wrap(f, wrapper2)

    f(1, 2, 3)

    assert channel1 == channel2 == [((1, 2, 3), {}), (1, 2, 3)]

    channel1[:] = []
    channel2[:] = []

    unwrap(f, wrapper1)
    f(1, 2, 3)

    assert not channel1 and channel2 == [((1, 2, 3), {}), (1, 2, 3)]

    channel2[:] = []

    unwrap(f, wrapper2)
    f(1, 2, 3)

    assert not channel1 and not channel2


def test_wrap_generator():
    channel = []

    def wrapper(f, args, kwargs):
        channel[:] = []
        for _ in f(*args, **kwargs):
            channel.append(_)
            yield _

    def g():
        for _ in range(10):
            yield _
        return

    wrap(g, wrapper)

    assert list(g()) == list(range(10)) == channel


def test_wrap_generator_send():
    def wrapper(f, args, kwargs):
        return f(*args, **kwargs)

    def g():
        yield 0
        for _ in range(1, 10):
            n = yield _
            assert _ == n
        return

    wrap(g, wrapper)

    gen = g()
    n = next(gen)
    channel = [n]
    try:
        while True:
            n = gen.send(n)
            channel.append(n)
    except StopIteration:
        pass

    assert list(range(10)) == channel


def test_wrap_generator_throw_close():
    def wrapper_maker(channel):
        def wrapper(f, args, kwargs):
            channel.append(True)

            __ddgen = f(*args, **kwargs)
            __ddgensend = __ddgen.send
            try:
                value = next(__ddgen)
                channel.append(value)
            except StopIteration:
                return
            while True:
                try:
                    tosend = yield value
                except GeneratorExit:
                    channel.append("GeneratorExit")
                    __ddgen.close()
                    raise GeneratorExit()
                except:  # noqa
                    channel.append(sys.exc_info()[0])
                    value = __ddgen.throw(*sys.exc_info())
                    channel.append(value)
                else:
                    try:
                        value = __ddgensend(tosend)
                        channel.append(value)
                    except StopIteration:
                        return

        return wrapper

    channel = []

    def g():
        while True:
            try:
                yield 0
            except ValueError:
                yield 1

    wrap(g, wrapper_maker(channel))
    inspect.isgeneratorfunction(g)

    gen = g()
    inspect.isgenerator(gen)

    for _ in range(10):
        assert next(gen) == 0
        assert gen.throw(ValueError) == 1

    gen.close()

    assert channel == [True] + [0, ValueError, 1] * 10 + ["GeneratorExit"]


def test_wrap_stack():
    def wrapper(f, args, kwargs):
        return f(*args, **kwargs)

    def f():
        stack = []
        frame = sys._getframe()
        while frame:
            stack.append(frame)
            frame = frame.f_back
        return stack

    wrap(f, wrapper)

    assert [frame.f_code.co_name for frame in f()[:4]] == ["f", "wrapper", "f", "test_wrap_stack"]



from ddtrace.internal.compat import ensure_pep562
from ddtrace.vendor.debtcollector import deprecate


# tags
CLUSTER = "cassandra.cluster"
KEYSPACE = "cassandra.keyspace"
CONSISTENCY_LEVEL = "cassandra.consistency_level"
PAGINATED = "cassandra.paginated"
PAGE_NUMBER = "cassandra.page_number"


def __getattr__(name):
    if name == "ROW_COUNT":
        deprecate(
            ("%s.%s is deprecated" % (__name__, name)),
            postfix=". Use ddtrace.ext.db.ROWCOUNT instead.",
            removal_version="2.0.0",
        )
        return "cassandra.row_count"

    if name in globals():
        return globals()[name]

    raise AttributeError("'%s' has no attribute '%s'", __name__, name)


ensure_pep562(__name__)



"""
Tracing utilities for the psycopg potgres client library.
"""
import functools

from psycopg2.extensions import connection
from psycopg2.extensions import cursor
from psycopg2.extensions import parse_dsn

from ddtrace import config
from ddtrace.internal.constants import COMPONENT

from ...constants import SPAN_KIND
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanKind
from ...ext import SpanTypes
from ...ext import db
from ...ext import net


class TracedCursor(cursor):
    """Wrapper around cursor creating one span per query"""

    def __init__(self, *args, **kwargs):
        self._datadog_tracer = kwargs.pop("datadog_tracer", None)
        self._datadog_service = kwargs.pop("datadog_service", None)
        self._datadog_tags = kwargs.pop("datadog_tags", None)
        super(TracedCursor, self).__init__(*args, **kwargs)

    def execute(self, query, vars=None):  # noqa: A002
        """just wrap the cursor execution in a span"""
        if not self._datadog_tracer:
            return cursor.execute(self, query, vars)

        with self._datadog_tracer.trace("postgres.query", service=self._datadog_service, span_type=SpanTypes.SQL) as s:
            s.set_tag_str(COMPONENT, config.psycopg.integration_name)
            s.set_tag_str(db.SYSTEM, config.psycopg.dbms_name)

            # set span.kind to the type of operation being performed
            s.set_tag_str(SPAN_KIND, SpanKind.CLIENT)

            s.set_tag(SPAN_MEASURED_KEY)
            if not s.sampled:
                return super(TracedCursor, self).execute(query, vars)

            s.resource = query
            s.set_tags(self._datadog_tags)
            try:
                return super(TracedCursor, self).execute(query, vars)
            finally:
                s.set_metric(db.ROWCOUNT, self.rowcount)

    def callproc(self, procname, vars=None):  # noqa: A002
        """just wrap the execution in a span"""
        return cursor.callproc(self, procname, vars)


class TracedConnection(connection):
    """Wrapper around psycopg2 for tracing"""

    def __init__(self, *args, **kwargs):

        self._datadog_tracer = kwargs.pop("datadog_tracer", None)
        self._datadog_service = kwargs.pop("datadog_service", None)

        super(TracedConnection, self).__init__(*args, **kwargs)

        # add metadata (from the connection, string, etc)
        dsn = parse_dsn(self.dsn)
        self._datadog_tags = {
            net.TARGET_HOST: dsn.get("host"),
            net.TARGET_PORT: dsn.get("port"),
            db.NAME: dsn.get("dbname"),
            db.USER: dsn.get("user"),
            db.SYSTEM: config.psycopg.dbms_name,
            "db.application": dsn.get("application_name"),
        }

        self._datadog_cursor_class = functools.partial(
            TracedCursor,
            datadog_tracer=self._datadog_tracer,
            datadog_service=self._datadog_service,
            datadog_tags=self._datadog_tags,
        )

    def cursor(self, *args, **kwargs):
        """register our custom cursor factory"""
        kwargs.setdefault("cursor_factory", self._datadog_cursor_class)
        return super(TracedConnection, self).cursor(*args, **kwargs)



import os
import sys
from typing import TYPE_CHECKING

from ddtrace.appsec.iast import oce
from ddtrace.appsec.iast._patch import set_and_check_module_is_patched
from ddtrace.appsec.iast._patch import set_module_unpatched
from ddtrace.appsec.iast._patch import try_unwrap
from ddtrace.appsec.iast._patch import try_wrap_function_wrapper
from ddtrace.appsec.iast.constants import DEFAULT_WEAK_HASH_ALGORITHMS
from ddtrace.appsec.iast.constants import EVIDENCE_ALGORITHM_TYPE
from ddtrace.appsec.iast.constants import MD5_DEF
from ddtrace.appsec.iast.constants import SHA1_DEF
from ddtrace.appsec.iast.constants import VULN_INSECURE_HASHING_TYPE
from ddtrace.appsec.iast.taint_sinks._base import VulnerabilityBase
from ddtrace.internal.logger import get_logger


if TYPE_CHECKING:  # pragma: no cover
    from typing import Any
    from typing import Callable
    from typing import Set

log = get_logger(__name__)


def get_weak_hash_algorithms():
    # type: () -> Set
    CONFIGURED_WEAK_HASH_ALGORITHMS = None
    DD_IAST_WEAK_HASH_ALGORITHMS = os.getenv("DD_IAST_WEAK_HASH_ALGORITHMS")
    if DD_IAST_WEAK_HASH_ALGORITHMS:
        CONFIGURED_WEAK_HASH_ALGORITHMS = set(algo.strip() for algo in DD_IAST_WEAK_HASH_ALGORITHMS.lower().split(","))

    return CONFIGURED_WEAK_HASH_ALGORITHMS or DEFAULT_WEAK_HASH_ALGORITHMS


@oce.register
class WeakHash(VulnerabilityBase):
    vulnerability_type = VULN_INSECURE_HASHING_TYPE
    evidence_type = EVIDENCE_ALGORITHM_TYPE


def unpatch_iast():
    # type: () -> None
    set_module_unpatched("hashlib", default_attr="_datadog_weak_hash_patch")
    set_module_unpatched("Crypto", default_attr="_datadog_weak_hash_patch")

    if sys.version_info >= (3, 0, 0):
        try_unwrap("_hashlib", "HASH.digest")
        try_unwrap("_hashlib", "HASH.hexdigest")
        try_unwrap(("_%s" % MD5_DEF), "MD5Type.digest")
        try_unwrap(("_%s" % MD5_DEF), "MD5Type.hexdigest")
        try_unwrap(("_%s" % SHA1_DEF), "SHA1Type.digest")
        try_unwrap(("_%s" % SHA1_DEF), "SHA1Type.hexdigest")
    else:
        try_unwrap("hashlib", MD5_DEF)
        try_unwrap("hashlib", SHA1_DEF)
        try_unwrap("hashlib", "new")

    # pycryptodome methods
    try_unwrap("Crypto.Hash.MD5", "MD5Hash.digest")
    try_unwrap("Crypto.Hash.MD5", "MD5Hash.hexdigest")
    try_unwrap("Crypto.Hash.SHA1", "SHA1Hash.digest")
    try_unwrap("Crypto.Hash.SHA1", "SHA1Hash.hexdigest")


def patch():
    # type: () -> None
    """Wrap hashing functions.
    Weak hashing algorithms are those that have been proven to be of high risk, or even completely broken,
    and thus are not fit for use.
    """

    if not set_and_check_module_is_patched("hashlib", default_attr="_datadog_weak_hash_patch"):
        return

    if not set_and_check_module_is_patched("Crypto", default_attr="_datadog_weak_hash_patch"):
        return

    weak_hash_algorithms = get_weak_hash_algorithms()

    if sys.version_info >= (3, 0, 0):
        try_wrap_function_wrapper("_hashlib", "HASH.digest", wrapped_digest_function)
        try_wrap_function_wrapper("_hashlib", "HASH.hexdigest", wrapped_digest_function)
        if MD5_DEF in weak_hash_algorithms:
            try_wrap_function_wrapper(("_%s" % MD5_DEF), "MD5Type.digest", wrapped_md5_function)
            try_wrap_function_wrapper(("_%s" % MD5_DEF), "MD5Type.hexdigest", wrapped_md5_function)
        if SHA1_DEF in weak_hash_algorithms:
            try_wrap_function_wrapper(("_%s" % SHA1_DEF), "SHA1Type.digest", wrapped_sha1_function)
            try_wrap_function_wrapper(("_%s" % SHA1_DEF), "SHA1Type.hexdigest", wrapped_sha1_function)
    else:
        if MD5_DEF in weak_hash_algorithms:
            try_wrap_function_wrapper("hashlib", MD5_DEF, wrapped_md5_function)
        if SHA1_DEF in weak_hash_algorithms:
            try_wrap_function_wrapper("hashlib", SHA1_DEF, wrapped_sha1_function)
        try_wrap_function_wrapper("hashlib", "new", wrapped_new_function)

    # pycryptodome methods
    if MD5_DEF in weak_hash_algorithms:
        try_wrap_function_wrapper("Crypto.Hash.MD5", "MD5Hash.digest", wrapped_md5_function)
        try_wrap_function_wrapper("Crypto.Hash.MD5", "MD5Hash.hexdigest", wrapped_md5_function)
    if SHA1_DEF in weak_hash_algorithms:
        try_wrap_function_wrapper("Crypto.Hash.SHA1", "SHA1Hash.digest", wrapped_sha1_function)
        try_wrap_function_wrapper("Crypto.Hash.SHA1", "SHA1Hash.hexdigest", wrapped_sha1_function)


@WeakHash.wrap
def wrapped_digest_function(wrapped, instance, args, kwargs):
    # type: (Callable, Any, Any, Any) -> Any
    if instance.name.lower() in get_weak_hash_algorithms():
        WeakHash.report(
            evidence_value=instance.name,
        )
    return wrapped(*args, **kwargs)


@WeakHash.wrap
def wrapped_md5_function(wrapped, instance, args, kwargs):
    # type: (Callable, Any, Any, Any) -> Any
    return wrapped_function(wrapped, MD5_DEF, instance, args, kwargs)


@WeakHash.wrap
def wrapped_sha1_function(wrapped, instance, args, kwargs):
    # type: (Callable, Any, Any, Any) -> Any
    return wrapped_function(wrapped, SHA1_DEF, instance, args, kwargs)


@WeakHash.wrap
def wrapped_new_function(wrapped, instance, args, kwargs):
    # type: (Callable, Any, Any, Any) -> Any
    if args[0].lower() in get_weak_hash_algorithms():
        WeakHash.report(
            evidence_value=args[0].lower(),
        )
    return wrapped(*args, **kwargs)


def wrapped_function(wrapped, evidence, instance, args, kwargs):
    # type: (Callable, str, Any, Any, Any) -> Any
    WeakHash.report(
        evidence_value=evidence,
    )
    return wrapped(*args, **kwargs)



import sys

import pytest

from ddtrace.appsec.iast.constants import EVIDENCE_ALGORITHM_TYPE
from ddtrace.appsec.iast.constants import VULN_INSECURE_HASHING_TYPE
from ddtrace.appsec.iast.taint_sinks.weak_hash import unpatch_iast
from ddtrace.constants import IAST_CONTEXT_KEY
from ddtrace.internal import _context
from tests.appsec.iast.fixtures.weak_algorithms import hashlib_new
from tests.appsec.iast.fixtures.weak_algorithms import parametrized_week_hash


FIXTURES_PATH = "tests/appsec/iast/fixtures/weak_algorithms.py"


@pytest.mark.parametrize(
    "hash_func,method", [("md5", "hexdigest"), ("md5", "digest"), ("sha1", "digest"), ("sha1", "hexdigest")]
)
def test_weak_hash_hashlib(iast_span_defaults, hash_func, method):
    parametrized_week_hash(hash_func, method)

    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)
    assert list(span_report.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report.vulnerabilities)[0].location.path.endswith(FIXTURES_PATH)
    assert list(span_report.vulnerabilities)[0].location.line == 14 if sys.version_info > (3, 0, 0) else 11
    assert list(span_report.vulnerabilities)[0].evidence.value == hash_func
    assert list(span_report.vulnerabilities)[0].evidence.type == EVIDENCE_ALGORITHM_TYPE


@pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="Digest is wrapped in Python 3")
@pytest.mark.parametrize("hash_func", ["md5", "sha1"])
def test_weak_hash_hashlib_no_digest(iast_span_md5_and_sha1_configured, hash_func):
    import hashlib

    m = getattr(hashlib, hash_func)()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")

    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_md5_and_sha1_configured)
    assert span_report is None


@pytest.mark.parametrize("hash_func,method", [("sha256", "digest"), ("sha256", "hexdigest")])
def test_weak_hash_secure_hash(iast_span_md5_and_sha1_configured, hash_func, method):
    import hashlib

    m = getattr(hashlib, hash_func)()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    getattr(m, method)()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_md5_and_sha1_configured)
    assert span_report is None


def test_weak_hash_new(iast_span_defaults):
    hashlib_new()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)

    assert list(span_report.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report.vulnerabilities)[0].location.path.endswith(FIXTURES_PATH)
    assert list(span_report.vulnerabilities)[0].location.line == 23 if sys.version_info > (3, 0, 0) else 20
    assert list(span_report.vulnerabilities)[0].evidence.value == "md5"
    assert list(span_report.vulnerabilities)[0].evidence.type == EVIDENCE_ALGORITHM_TYPE


def test_weak_hash_new_with_child_span(tracer, iast_span_defaults):
    with tracer.trace("test_child") as span:
        hashlib_new()
        span_report1 = _context.get_item(IAST_CONTEXT_KEY, span=span)

    span_report2 = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)

    assert list(span_report1.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report1.vulnerabilities)[0].location.path.endswith(FIXTURES_PATH)
    assert list(span_report1.vulnerabilities)[0].evidence.value == "md5"

    assert list(span_report2.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report2.vulnerabilities)[0].location.path.endswith(FIXTURES_PATH)
    assert list(span_report2.vulnerabilities)[0].evidence.value == "md5"


@pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="_md5 works only in Python 3")
def test_weak_hash_md5_builtin_py3_unpatched(iast_span_md5_and_sha1_configured):
    import _md5

    unpatch_iast()
    m = _md5.md5()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_md5_and_sha1_configured)

    assert span_report is None


@pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="_md5 works only in Python 3")
def test_weak_hash_md5_builtin_py3_md5_and_sha1_configured(iast_span_defaults):
    import _md5

    m = _md5.md5()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)

    assert list(span_report.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report.vulnerabilities)[0].location.path.endswith("tests/appsec/iast/test_weak_hash.py")
    assert list(span_report.vulnerabilities)[0].evidence.value == "md5"


@pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="_md5 works only in Python 3")
def test_weak_hash_md5_builtin_py3_only_md4_configured(iast_span_only_md4):
    import _md5

    m = _md5.md5()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_only_md4)

    assert span_report is None


@pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="_md5 works only in Python 3")
def test_weak_hash_md5_builtin_py3_only_md5_configured(iast_span_only_md5):
    import _md5

    m = _md5.md5()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_only_md5)

    assert list(span_report.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report.vulnerabilities)[0].location.path.endswith("tests/appsec/iast/test_weak_hash.py")
    assert list(span_report.vulnerabilities)[0].evidence.value == "md5"


@pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="_md5 works only in Python 3")
def test_weak_hash_md5_builtin_py3_only_sha1_configured(iast_span_only_sha1):
    import _md5

    m = _md5.md5()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_only_sha1)

    assert span_report is None


@pytest.mark.skipif(sys.version_info > (3, 0, 0), reason="md5 works only in Python 2")
def test_weak_hash_md5_builtin_py2(iast_span_defaults):
    import md5

    m = md5.md5()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)
    assert list(span_report.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report.vulnerabilities)[0].location.path.endswith("tests/appsec/iast/test_weak_hash.py")
    assert list(span_report.vulnerabilities)[0].evidence.value == "md5"


def test_weak_hash_pycryptodome_hashes_md5(iast_span_defaults):
    from Crypto.Hash import MD5

    m = MD5.new()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)
    assert list(span_report.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report.vulnerabilities)[0].location.path.endswith("tests/appsec/iast/test_weak_hash.py")
    assert list(span_report.vulnerabilities)[0].evidence.value == "md5"


def test_weak_hash_pycryptodome_hashes_sha1_defaults(iast_span_defaults):
    from Crypto.Hash import SHA1

    m = SHA1.new()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)

    assert list(span_report.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report.vulnerabilities)[0].location.path.endswith("tests/appsec/iast/test_weak_hash.py")
    assert list(span_report.vulnerabilities)[0].evidence.value == "sha1"


def test_weak_hash_pycryptodome_hashes_sha1_only_md5_configured(iast_span_only_md5):
    from Crypto.Hash import SHA1

    m = SHA1.new()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_only_md5)

    assert span_report is None


def test_weak_hash_pycryptodome_hashes_sha1_only_sha1_configured(iast_span_only_sha1):
    from Crypto.Hash import SHA1

    m = SHA1.new()
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_only_sha1)

    assert list(span_report.vulnerabilities)[0].type == VULN_INSECURE_HASHING_TYPE
    assert list(span_report.vulnerabilities)[0].location.path.endswith("tests/appsec/iast/test_weak_hash.py")
    assert list(span_report.vulnerabilities)[0].evidence.value == "sha1"


def test_weak_check_repeated(iast_span_defaults):
    import hashlib

    m = hashlib.new("md5")
    m.update(b"Nobody inspects")
    m.update(b" the spammish repetition")
    num_vulnerabilities = 10
    for i in range(0, num_vulnerabilities):
        m.digest()

    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)

    assert len(span_report.vulnerabilities) == 1



import abc
import base64
from datetime import datetime
import hashlib
import json
import re
import sys
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import Set
from typing import TYPE_CHECKING
import uuid

import attr
import cattr
import six

import ddtrace
from ddtrace.appsec.utils import _appsec_rc_capabilities
from ddtrace.internal import agent
from ddtrace.internal import runtime
from ddtrace.internal.hostname import get_hostname
from ddtrace.internal.logger import get_logger
from ddtrace.internal.remoteconfig.constants import REMOTE_CONFIG_AGENT_ENDPOINT
from ddtrace.internal.runtime import container
from ddtrace.internal.utils.time import parse_isoformat

from ..utils.version import _pep440_to_semver


if TYPE_CHECKING:  # pragma: no cover
    from typing import MutableMapping
    from typing import Tuple
    from typing import Union


log = get_logger(__name__)

TARGET_FORMAT = re.compile(r"^(datadog/\d+|employee)/([^/]+)/([^/]+)/([^/]+)$")


class RemoteConfigError(Exception):
    """
    An error occurred during the configuration update procedure.
    The error is reported to the agent.
    """


@attr.s(frozen=True)
class ConfigMetadata(object):
    """
    Configuration TUF target metadata
    """

    id = attr.ib(type=str)
    product_name = attr.ib(type=str)
    sha256_hash = attr.ib(type=Optional[str])
    length = attr.ib(type=Optional[int])
    tuf_version = attr.ib(type=Optional[int])


@attr.s
class Signature(object):
    keyid = attr.ib(type=str)
    sig = attr.ib(type=str)


@attr.s
class Key(object):
    keytype = attr.ib(type=str)
    keyid_hash_algorithms = attr.ib(type=List[str])
    keyval = attr.ib(type=Mapping)
    scheme = attr.ib(type=str)


@attr.s
class Role(object):
    keyids = attr.ib(type=List[str])
    threshold = attr.ib(type=int)


@attr.s
class Root(object):
    _type = attr.ib(type=str, validator=attr.validators.in_(("root",)))
    spec_version = attr.ib(type=str)
    consistent_snapshot = attr.ib(type=bool)
    expires = attr.ib(type=datetime, converter=parse_isoformat)
    keys = attr.ib(type=Mapping[str, Key])
    roles = attr.ib(type=Mapping[str, Role])
    version = attr.ib(type=int)


@attr.s
class SignedRoot(object):
    signatures = attr.ib(type=List[Signature])
    signed = attr.ib(type=Root)


@attr.s
class TargetDesc(object):
    length = attr.ib(type=int)
    hashes = attr.ib(type=Mapping[str, str])
    custom = attr.ib(type=Mapping[str, Any])


@attr.s
class Targets(object):
    _type = attr.ib(type=str, validator=attr.validators.in_(("targets",)))
    custom = attr.ib(type=Mapping[str, Any])
    expires = attr.ib(type=datetime, converter=parse_isoformat)
    spec_version = attr.ib(type=str, validator=attr.validators.in_(("1.0", "1.0.0")))
    targets = attr.ib(type=Mapping[str, TargetDesc])
    version = attr.ib(type=int)


@attr.s
class SignedTargets(object):
    signatures = attr.ib(type=List[Signature])
    signed = attr.ib(type=Targets)


@attr.s
class TargetFile(object):
    path = attr.ib(type=str)
    raw = attr.ib(type=str)


@attr.s
class AgentPayload(object):
    roots = attr.ib(type=List[SignedRoot], default=None)
    targets = attr.ib(type=SignedTargets, default=None)
    target_files = attr.ib(type=List[TargetFile], default=[])
    client_configs = attr.ib(type=Set[str], default={})


def _load_json(data):
    # type: (Union[str, bytes]) -> Dict[str, Any]
    if (3, 6) > sys.version_info > (3,) and isinstance(data, six.binary_type):
        data = str(data, encoding="utf-8")
    return json.loads(data)


AppliedConfigType = Dict[str, ConfigMetadata]
TargetsType = Dict[str, ConfigMetadata]


class RemoteConfigCallBack(six.with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def __call__(self, metadata, config):
        # type: (Optional[ConfigMetadata], Any) -> None
        pass


class RemoteConfigCallBackAfterMerge(six.with_metaclass(abc.ABCMeta)):
    configs = {}  # type: Dict[str, Any]

    @abc.abstractmethod
    def __call__(self, target, config):
        # type: (str, Any) -> None
        pass

    def append(self, target, config):
        if not self.configs.get(target):
            self.configs[target] = {}
        if config is False:
            # Remove old config from the configs dict. _remove_previously_applied_configurations function should
            # call to this method
            del self.configs[target]
        elif config is not None:
            # Append the new config to the configs dict. _load_new_configurations function should
            # call to this method
            if isinstance(config, dict):
                self.configs[target].update(config)
            else:
                raise ValueError("target %s config %s has type of %s" % (target, config, type(config)))

    def dispatch(self):
        config_result = {}
        for target, config in self.configs.items():
            for key, value in config.items():
                if isinstance(value, list):
                    config_result[key] = config_result.get(key, []) + value
                else:
                    raise ValueError("target %s key %s has type of %s" % (target, key, type(value)))
        if config_result:
            self.__call__("", config_result)


class RemoteConfigClient(object):
    """
    The Remote Configuration client regularly checks for updates on the agent
    and dispatches configurations to registered products.
    """

    def __init__(self):
        # type: () -> None
        tracer_version = _pep440_to_semver()

        self.id = str(uuid.uuid4())
        self.agent_url = agent_url = agent.get_trace_url()
        self._conn = agent.get_connection(agent_url, timeout=agent.get_trace_agent_timeout())
        self._headers = {"content-type": "application/json"}

        container_info = container.get_container_info()
        if container_info is not None:
            container_id = container_info.container_id
            if container_id is not None:
                self._headers["Datadog-Container-Id"] = container_id

        tags = ddtrace.config.tags.copy()
        if ddtrace.config.env:
            tags["env"] = ddtrace.config.env
        if ddtrace.config.version:
            tags["version"] = ddtrace.config.version
        tags["tracer_version"] = tracer_version
        tags["host_name"] = get_hostname()

        self._client_tracer = dict(
            runtime_id=runtime.get_runtime_id(),
            language="python",
            tracer_version=tracer_version,
            service=ddtrace.config.service,
            env=ddtrace.config.env,
            app_version=ddtrace.config.version,
            tags=[":".join(_) for _ in tags.items()],
        )
        self.cached_target_files = []  # type: List[AppliedConfigType]
        self.converter = cattr.Converter()

        # cattrs doesn't implement datetime converter in Py27, we should register
        def date_to_fromisoformat(val, cls):
            return val

        self.converter.register_structure_hook(datetime, date_to_fromisoformat)

        def base64_to_struct(val, cls):
            raw = base64.b64decode(val)
            obj = _load_json(raw)
            return self.converter.structure_attrs_fromdict(obj, cls)

        self.converter.register_structure_hook(SignedRoot, base64_to_struct)
        self.converter.register_structure_hook(SignedTargets, base64_to_struct)

        self._products = dict()  # type: MutableMapping[str, RemoteConfigCallBack]
        self._applied_configs = dict()  # type: AppliedConfigType
        self._last_targets_version = 0
        self._last_error = None  # type: Optional[str]
        self._backend_state = None  # type: Optional[str]

    def register_product(self, product_name, func=None):
        # type: (str, Optional[RemoteConfigCallBack]) -> None
        if func is not None:
            self._products[product_name] = func
        else:
            self._products.pop(product_name, None)

    def unregister_product(self, product_name):
        # type: (str) -> None
        self._products.pop(product_name, None)

    def _send_request(self, payload):
        # type: (str) -> Optional[Mapping[str, Any]]
        try:
            self._conn.request("POST", REMOTE_CONFIG_AGENT_ENDPOINT, payload, self._headers)
            resp = self._conn.getresponse()
            data = resp.read()
        except OSError as e:
            log.warning("Unexpected connection error in remote config client request: %s", str(e))  # noqa: G200
            return None
        finally:
            self._conn.close()

        if resp.status == 404:
            # Remote configuration is not enabled or unsupported by the agent
            return None

        if resp.status < 200 or resp.status >= 300:
            log.warning("Unexpected error: HTTP error status %s, reason %s", resp.status, resp.reason)
            return None

        return json.loads(data)

    @staticmethod
    def _extract_target_file(payload, target, config):
        # type: (AgentPayload, str, ConfigMetadata) -> Optional[Dict[str, Any]]
        candidates = [item.raw for item in payload.target_files if item.path == target]
        if len(candidates) != 1 or candidates[0] is None:
            log.debug(
                "invalid target_files for %r. target files: %s", target, [item.path for item in payload.target_files]
            )
            return None

        try:
            raw = base64.b64decode(candidates[0])
        except Exception:
            raise RemoteConfigError("invalid base64 target_files for {!r}".format(target))

        computed_hash = hashlib.sha256(raw).hexdigest()
        if computed_hash != config.sha256_hash:
            raise RemoteConfigError(
                "mismatch between target {!r} hashes {!r} != {!r}".format(target, computed_hash, config.sha256_hash)
            )

        try:
            return _load_json(raw)
        except Exception:
            raise RemoteConfigError("invalid JSON content for target {!r}".format(target))

    @staticmethod
    def _parse_target(target, metadata):
        # type: (str, TargetDesc) -> ConfigMetadata
        m = TARGET_FORMAT.match(target)
        if m is None:
            raise RemoteConfigError("unexpected target format {!r}".format(target))
        _, product_name, config_id, _ = m.groups()
        return ConfigMetadata(
            id=config_id,
            product_name=product_name,
            sha256_hash=metadata.hashes.get("sha256"),
            length=metadata.length,
            tuf_version=metadata.custom.get("v"),
        )

    def _build_payload(self, state):
        # type: (Mapping[str, Any]) -> Mapping[str, Any]
        return dict(
            client=dict(
                id=self.id,
                products=list(self._products.keys()),
                is_tracer=True,
                client_tracer=self._client_tracer,
                state=state,
                capabilities=_appsec_rc_capabilities(),
            ),
            cached_target_files=self.cached_target_files,
        )

    def _build_state(self):
        # type: () -> Mapping[str, Any]
        has_error = self._last_error is not None
        state = dict(
            root_version=1,
            targets_version=self._last_targets_version,
            config_states=[
                dict(id=config.id, version=config.tuf_version, product=config.product_name)
                for config in self._applied_configs.values()
            ],
            has_error=has_error,
        )
        if self._backend_state is not None:
            state["backend_client_state"] = self._backend_state
        if has_error:
            state["error"] = self._last_error
        return state

    def _process_targets(self, payload):
        # type: (AgentPayload) -> Tuple[Optional[int], Optional[str], Optional[TargetsType]]
        if payload.targets is None:
            # no targets received
            return None, None, None

        signed = payload.targets.signed
        targets = dict()  # type: TargetsType

        for target, metadata in signed.targets.items():
            config = self._parse_target(target, metadata)
            if config is not None:
                targets[target] = config

        backend_state = signed.custom.get("opaque_backend_state")
        return signed.version, backend_state, targets

    @staticmethod
    def _apply_callback(list_callbacks, callback, config_content, target, config):
        # type: (List[RemoteConfigCallBackAfterMerge], Any, Any, str, ConfigMetadata) -> None
        if isinstance(callback, RemoteConfigCallBackAfterMerge):
            callback.append(target, config_content)
            if callback not in list_callbacks and not any(
                filter(lambda x: isinstance(x, type(callback)), list_callbacks)
            ):
                list_callbacks.append(callback)
        else:
            callback(config, config_content)

    def _remove_previously_applied_configurations(self, applied_configs, client_configs, targets):
        # type: (AppliedConfigType, TargetsType, TargetsType) -> None
        list_callbacks = []  # type: List[RemoteConfigCallBackAfterMerge]
        for target, config in self._applied_configs.items():
            if target in client_configs and targets.get(target) == config:
                # The configuration has not changed.
                applied_configs[target] = config
                continue
            elif target not in client_configs:
                log.debug("Disable configuration: %s", target)
                callback_action = False
            else:
                continue

            callback = self._products.get(config.product_name)
            if callback:
                try:
                    log.debug("Disable configuration 2: %s. ", target)
                    self._apply_callback(list_callbacks, callback, callback_action, target, config)
                except Exception:
                    log.debug("error while removing product %s config %r", config.product_name, config)
                    continue

        for callback_to_dispach in list_callbacks:
            callback_to_dispach.dispatch()

    def _load_new_configurations(self, applied_configs, client_configs, payload):
        # type: (AppliedConfigType, TargetsType, AgentPayload) -> None
        list_callbacks = []  # type: List[RemoteConfigCallBackAfterMerge]
        for target, config in client_configs.items():
            callback = self._products.get(config.product_name)
            if callback:
                applied_config = self._applied_configs.get(target)
                if applied_config == config:
                    continue

                config_content = self._extract_target_file(payload, target, config)
                if config_content is None:
                    continue

                try:
                    log.debug("Load new configuration: %s. content ", target)
                    self._apply_callback(list_callbacks, callback, config_content, target, config)
                except Exception:
                    log.debug(
                        "Failed to load configuration %s for product %r", config, config.product_name, exc_info=True
                    )
                    continue
                else:
                    applied_configs[target] = config

        for callback_to_dispach in list_callbacks:
            callback_to_dispach.dispatch()

    def _add_apply_config_to_cache(self):
        if self._applied_configs:
            cached_data = []
            for target, config in self._applied_configs.items():
                cached_data.append(
                    {
                        "path": target,
                        "length": config.length,
                        "hashes": [{"algorithm": "sha256", "hash": config.sha256_hash}],
                    }
                )
            self.cached_target_files = cached_data
        else:
            self.cached_target_files = []

    def _validate_config_exists_in_target_paths(self, payload_client_configs, payload_target_files):
        # type: (Set[str], List[TargetFile]) -> None
        paths = {_.path for _ in payload_target_files}
        paths = paths.union({_["path"] for _ in self.cached_target_files})

        # !(payload.client_configs is a subset of paths or payload.client_configs is equal to paths)
        if not set(payload_client_configs) <= paths:
            raise RemoteConfigError("Not all client configurations have target files")

    @staticmethod
    def _validate_signed_target_files(payload_target_files, payload_targets_signed, client_configs):
        # type: (List[TargetFile], Targets, TargetsType) -> None
        for target in payload_target_files:
            if (payload_targets_signed.targets and not payload_targets_signed.targets.get(target.path)) and (
                client_configs and not client_configs.get(target.path)
            ):
                raise RemoteConfigError(
                    "target file %s not exists in client_config and signed targets" % (target.path,)
                )

    def _process_response(self, data):
        # type: (Mapping[str, Any]) -> None
        try:
            # log.debug("response payload: %r", data)
            payload = self.converter.structure_attrs_fromdict(data, AgentPayload)
        except Exception:
            log.debug("invalid agent payload received: %r", data, exc_info=True)
            raise RemoteConfigError("invalid agent payload received")

        self._validate_config_exists_in_target_paths(payload.client_configs, payload.target_files)

        # 1. Deserialize targets
        last_targets_version, backend_state, targets = self._process_targets(payload)
        if last_targets_version is None or targets is None:
            log.debug("No targets in configuration payload")
            for callback in self._products.values():
                callback(None, None)
            return

        client_configs = {k: v for k, v in targets.items() if k in payload.client_configs}
        log.debug("Retrieved client configs last version %s: %s", last_targets_version, client_configs)

        self._validate_signed_target_files(payload.target_files, payload.targets.signed, client_configs)

        # 2. Remove previously applied configurations
        applied_configs = dict()  # type: AppliedConfigType
        self._remove_previously_applied_configurations(applied_configs, client_configs, targets)

        # 3. Load new configurations
        self._load_new_configurations(applied_configs, client_configs, payload)

        self._last_targets_version = last_targets_version
        self._applied_configs = applied_configs
        self._backend_state = backend_state

        self._add_apply_config_to_cache()

    def request(self):
        # type: () -> bool
        try:
            state = self._build_state()
            payload = json.dumps(self._build_payload(state))
            response = self._send_request(payload)
            if response is None:
                return False
            self._process_response(response)
            self._last_error = None
            return True

        except RemoteConfigError as e:
            self._last_error = str(e)
            log.debug("remote configuration client reported an error", exc_info=True)
        except ValueError:
            log.debug("Unexpected response data", exc_info=True)
        except Exception:
            log.debug("Unexpected error", exc_info=True)

        return False



"""
The gevent integration adds support for tracing across greenlets.


The integration patches the gevent internals to add context management logic.

.. note::
    If ``ddtrace-run`` is not being used then be sure to ``import ddtrace.auto``
    before calling ``gevent.monkey.patch_all``.
    If ``ddtrace-run`` is being used then no additional configuration is required.


The integration also configures the global tracer instance to use a gevent
context provider to utilize the context management logic.

If custom tracer instances are being used in a gevent application, then
configure it with::

    from ddtrace.contrib.gevent import context_provider

    tracer.configure(context_provider=context_provider)


Enabling
~~~~~~~~

The integration is enabled automatically when using
:ref:`ddtrace-run<ddtracerun>` or :func:`patch_all()<ddtrace.patch_all>`.

Or use :func:`patch()<ddtrace.patch>` to manually enable the integration::

    from ddtrace import patch
    patch(gevent=True)


Example of the context propagation::

    def my_parent_function():
        with tracer.trace("web.request") as span:
            span.service = "web"
            gevent.spawn(worker_function)


    def worker_function():
        # then trace its child
        with tracer.trace("greenlet.call") as span:
            span.service = "greenlet"
            ...

            with tracer.trace("greenlet.child_call") as child:
                ...
"""
from ...internal.utils.importlib import require_modules


required_modules = ["gevent"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch
        from .patch import unpatch
        from .provider import GeventContextProvider

        context_provider = GeventContextProvider()

        __all__ = [
            "patch",
            "unpatch",
            "context_provider",
        ]



from contextlib import contextmanager
import random

import pylibmc

# project
import ddtrace
from ddtrace import config
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.constants import SPAN_KIND
from ddtrace.constants import SPAN_MEASURED_KEY
from ddtrace.contrib.pylibmc.addrs import parse_addresses
from ddtrace.ext import SpanKind
from ddtrace.ext import SpanTypes
from ddtrace.ext import db
from ddtrace.ext import memcached
from ddtrace.ext import net
from ddtrace.internal.compat import Iterable
from ddtrace.internal.constants import COMPONENT
from ddtrace.internal.logger import get_logger
from ddtrace.vendor.wrapt import ObjectProxy


# Original Client class
_Client = pylibmc.Client


log = get_logger(__name__)


class TracedClient(ObjectProxy):
    """TracedClient is a proxy for a pylibmc.Client that times it's network operations."""

    def __init__(self, client=None, service=memcached.SERVICE, tracer=None, *args, **kwargs):
        """Create a traced client that wraps the given memcached client."""
        # The client instance/service/tracer attributes are kept for compatibility
        # with the old interface: TracedClient(client=pylibmc.Client(['localhost:11211']))
        # TODO(Benjamin): Remove these in favor of patching.
        if not isinstance(client, _Client):
            # We are in the patched situation, just pass down all arguments to the pylibmc.Client
            # Note that, in that case, client isn't a real client (just the first argument)
            client = _Client(client, *args, **kwargs)
        else:
            log.warning(
                "TracedClient instantiation is deprecated and will be remove "
                "in future versions (0.6.0). Use patching instead (see the docs)."
            )

        super(TracedClient, self).__init__(client)

        pin = ddtrace.Pin(service=service, tracer=tracer)
        pin.onto(self)

        # attempt to collect the pool of urls this client talks to
        try:
            self._addresses = parse_addresses(client.addresses)
        except Exception:
            log.debug("error setting addresses", exc_info=True)

    def clone(self, *args, **kwargs):
        # rewrap new connections.
        cloned = self.__wrapped__.clone(*args, **kwargs)
        traced_client = TracedClient(cloned)
        pin = ddtrace.Pin.get_from(self)
        if pin:
            pin.clone().onto(traced_client)
        return traced_client

    def get(self, *args, **kwargs):
        return self._trace_cmd("get", *args, **kwargs)

    def set(self, *args, **kwargs):
        return self._trace_cmd("set", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self._trace_cmd("delete", *args, **kwargs)

    def gets(self, *args, **kwargs):
        return self._trace_cmd("gets", *args, **kwargs)

    def touch(self, *args, **kwargs):
        return self._trace_cmd("touch", *args, **kwargs)

    def cas(self, *args, **kwargs):
        return self._trace_cmd("cas", *args, **kwargs)

    def incr(self, *args, **kwargs):
        return self._trace_cmd("incr", *args, **kwargs)

    def decr(self, *args, **kwargs):
        return self._trace_cmd("decr", *args, **kwargs)

    def append(self, *args, **kwargs):
        return self._trace_cmd("append", *args, **kwargs)

    def prepend(self, *args, **kwargs):
        return self._trace_cmd("prepend", *args, **kwargs)

    def get_multi(self, *args, **kwargs):
        return self._trace_multi_cmd("get_multi", *args, **kwargs)

    def set_multi(self, *args, **kwargs):
        return self._trace_multi_cmd("set_multi", *args, **kwargs)

    def delete_multi(self, *args, **kwargs):
        return self._trace_multi_cmd("delete_multi", *args, **kwargs)

    def _trace_cmd(self, method_name, *args, **kwargs):
        """trace the execution of the method with the given name and will
        patch the first arg.
        """
        method = getattr(self.__wrapped__, method_name)
        with self._span(method_name) as span:

            if span and args:
                span.set_tag_str(memcached.QUERY, "%s %s" % (method_name, args[0]))

            if method_name == "get":
                result = method(*args, **kwargs)
                span.set_metric(db.ROWCOUNT, 1 if result else 0)
                return result
            elif method_name == "gets":
                result = method(*args, **kwargs)

                # returns a tuple object that may be (None, None)
                span.set_metric(db.ROWCOUNT, 1 if isinstance(result, Iterable) and len(result) > 0 and result[0] else 0)
                return result
            else:
                return method(*args, **kwargs)

    def _trace_multi_cmd(self, method_name, *args, **kwargs):
        """trace the execution of the multi command with the given name."""
        method = getattr(self.__wrapped__, method_name)
        with self._span(method_name) as span:

            pre = kwargs.get("key_prefix")
            if span and pre:
                span.set_tag_str(memcached.QUERY, "%s %s" % (method_name, pre))

            if method_name == "get_multi":
                result = method(*args, **kwargs)

                # returns mapping of key -> value if key exists, but does not include a missing key. Empty result = {}
                span.set_metric(
                    db.ROWCOUNT, sum(1 for doc in result if doc) if result and isinstance(result, Iterable) else 0
                )
                return result
            else:
                return method(*args, **kwargs)

    @contextmanager
    def _no_span(self):
        yield None

    def _span(self, cmd_name):
        """Return a span timing the given command."""
        pin = ddtrace.Pin.get_from(self)
        if not pin or not pin.enabled():
            return self._no_span()

        span = pin.tracer.trace(
            "memcached.cmd",
            service=pin.service,
            resource=cmd_name,
            span_type=SpanTypes.CACHE,
        )

        span.set_tag_str(COMPONENT, config.pylibmc.integration_name)
        span.set_tag_str(db.SYSTEM, memcached.DBMS_NAME)

        # set span.kind to the type of operation being performed
        span.set_tag_str(SPAN_KIND, SpanKind.CLIENT)

        span.set_tag(SPAN_MEASURED_KEY)

        try:
            self._tag_span(span)
        except Exception:
            log.debug("error tagging span", exc_info=True)
        return span

    def _tag_span(self, span):
        # FIXME[matt] the host selection is buried in c code. we can't tell what it's actually
        # using, so fallback to randomly choosing one. can we do better?
        if self._addresses:
            _, host, port, _ = random.choice(self._addresses)
            span.set_tag_str(net.TARGET_HOST, host)
            span.set_tag(net.TARGET_PORT, port)

        # set analytics sample rate
        span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, config.pylibmc.get_analytics_sample_rate())



import logging
import os

from ddtrace.internal import agent
from ddtrace.internal import periodic
from ddtrace.internal.logger import get_logger
from ddtrace.internal.remoteconfig.client import RemoteConfigClient
from ddtrace.internal.remoteconfig.constants import REMOTE_CONFIG_AGENT_ENDPOINT
from ddtrace.internal.utils.formats import asbool
from ddtrace.internal.utils.time import StopWatch
from ddtrace.vendor.debtcollector import deprecate


log = get_logger(__name__)


DEFAULT_REMOTECONFIG_POLL_SECONDS = 5.0  # seconds


def get_poll_interval_seconds():
    # type:() -> float
    if os.getenv("DD_REMOTECONFIG_POLL_SECONDS"):
        deprecate(
            "Using environment variable 'DD_REMOTECONFIG_POLL_SECONDS' is deprecated",
            message="Please use DD_REMOTE_CONFIG_POLL_INTERVAL_SECONDS instead.",
            removal_version="2.0.0",
        )
    return float(
        os.getenv(
            "DD_REMOTE_CONFIG_POLL_INTERVAL_SECONDS",
            default=os.getenv("DD_REMOTECONFIG_POLL_SECONDS", default=DEFAULT_REMOTECONFIG_POLL_SECONDS),
        )
    )


class RemoteConfigWorker(periodic.PeriodicService):
    """Remote configuration worker.

    This implements a finite-state machine that allows checking the agent for
    the expected endpoint, which could be enabled after the client is started.
    """

    def __init__(self):
        super(RemoteConfigWorker, self).__init__(interval=get_poll_interval_seconds())
        self._client = RemoteConfigClient()
        log.debug("RemoteConfigWorker created with polling interval %d", get_poll_interval_seconds())
        self._state = self._agent_check

    def _agent_check(self):
        # type: () -> None
        try:
            info = agent.info()
        except Exception:
            info = None

        if info:
            endpoints = info.get("endpoints", [])
            if endpoints and (
                REMOTE_CONFIG_AGENT_ENDPOINT in endpoints or ("/" + REMOTE_CONFIG_AGENT_ENDPOINT) in endpoints
            ):
                self._state = self._online
                return

        if asbool(os.environ.get("DD_TRACE_DEBUG")) or "DD_REMOTE_CONFIGURATION_ENABLED" in os.environ:
            LOG_LEVEL = logging.WARNING
        else:
            LOG_LEVEL = logging.DEBUG

        log.log(
            LOG_LEVEL,
            "Agent is down or Remote Config is not enabled in the Agent\n"
            "Check your Agent version, you need an Agent running on 7.39.1 version or above.\n"
            "Check Your Remote Config environment variables on your Agent:\n"
            "DD_REMOTE_CONFIGURATION_ENABLED=true\n"
            "See: https://docs.datadoghq.com/agent/guide/how_remote_config_works/",
        )

    def _online(self):
        # type: () -> None
        with StopWatch() as sw:
            if not self._client.request():
                # An error occurred, so we transition back to the agent check
                self._state = self._agent_check
                return

        elapsed = sw.elapsed()
        if elapsed >= self.interval:
            log_level = logging.WARNING
        else:
            log_level = logging.DEBUG
        log.log(log_level, "request config in %.5fs to %s", elapsed, self._client.agent_url)

    def periodic(self):
        # type: () -> None
        return self._state()



import typing

import attr

from ddtrace.ext import SpanTypes
from ddtrace.internal import forksafe
from ddtrace.internal.compat import ensure_str
from ddtrace.internal.processor import SpanProcessor
from ddtrace.span import Span


EndpointCountsType = typing.Dict[str, int]


@attr.s(eq=False)
class EndpointCallCounterProcessor(SpanProcessor):

    endpoint_counts = attr.ib(init=False, repr=False, type=EndpointCountsType, factory=lambda: {}, eq=False)
    _endpoint_counts_lock = attr.ib(init=False, repr=False, factory=forksafe.Lock, eq=False)
    _enabled = attr.ib(default=False, repr=False, eq=False)

    def enable(self):
        # type: () -> None
        self._enabled = True

    def on_span_start(self, span):
        # type: (Span) -> None
        pass

    def on_span_finish(self, span):
        # type: (Span) -> None
        if not self._enabled:
            return
        if span._local_root == span and span.span_type == SpanTypes.WEB:
            resource = ensure_str(span.resource, errors="backslashreplace")
            with self._endpoint_counts_lock:
                self.endpoint_counts[resource] = self.endpoint_counts.get(resource, 0) + 1

    def reset(self):
        # type: () -> EndpointCountsType
        with self._endpoint_counts_lock:
            counts = self.endpoint_counts
            self.endpoint_counts = {}
            return counts



import ctypes
import gc
import sys
from typing import TYPE_CHECKING

from ddtrace.internal.logger import get_logger
from ddtrace.vendor.wrapt import FunctionWrapper
from ddtrace.vendor.wrapt import resolve_path


if TYPE_CHECKING:  # pragma: no cover
    from typing import Any
    from typing import Callable
    from typing import Dict
    from typing import Optional


_DD_ORIGINAL_ATTRIBUTES = {}  # type: Dict[Any, Any]

log = get_logger(__name__)


def set_and_check_module_is_patched(module_str, default_attr="_datadog_patch"):
    # type: (str, str) -> Optional[bool]
    try:
        __import__(module_str)
        module = sys.modules[module_str]
        if getattr(module, default_attr, False):
            return False
        setattr(module, default_attr, True)
    except ImportError:
        pass
    return True


def set_module_unpatched(module_str, default_attr="_datadog_patch"):
    # type: (str, str) -> None
    try:
        __import__(module_str)
        module = sys.modules[module_str]
        setattr(module, default_attr, False)
    except ImportError:
        pass


def try_wrap_function_wrapper(module, name, wrapper):
    # type: (str, str, Callable) -> None
    try:
        wrap_object(module, name, FunctionWrapper, (wrapper,))
    except (ImportError, AttributeError):
        log.debug("IAST patching. Module %s.%s not exists", module, name)


def try_unwrap(module, name):
    (parent, attribute, _) = resolve_path(module, name)
    if (parent, attribute) in _DD_ORIGINAL_ATTRIBUTES:
        original = _DD_ORIGINAL_ATTRIBUTES[(parent, attribute)]
        apply_patch(parent, attribute, original)
        del _DD_ORIGINAL_ATTRIBUTES[(parent, attribute)]


def apply_patch(parent, attribute, replacement):
    try:
        current_attribute = getattr(parent, attribute)
        # Avoid overwriting the original function if we call this twice
        if not isinstance(current_attribute, FunctionWrapper):
            _DD_ORIGINAL_ATTRIBUTES[(parent, attribute)] = current_attribute
        setattr(parent, attribute, replacement)
    except (TypeError, AttributeError):
        patch_builtins(parent, attribute, replacement)


def wrap_object(module, name, factory, args=(), kwargs={}):
    (parent, attribute, original) = resolve_path(module, name)
    wrapper = factory(original, *args, **kwargs)
    apply_patch(parent, attribute, wrapper)
    return wrapper


def patchable_builtin(klass):
    refs = gc.get_referents(klass.__dict__)
    assert len(refs) == 1
    return refs[0]


def patch_builtins(klass, attr, value):
    """Based on forbiddenfruit package:
    https://github.com/clarete/forbiddenfruit/blob/master/forbiddenfruit/__init__.py#L421
    ---
    Patch a built-in `klass` with `attr` set to `value`

    This function monkey-patches the built-in python object `attr` adding a new
    attribute to it. You can add any kind of argument to the `class`.

    It's possible to attach methods as class methods, just do the following:

      >>> def myclassmethod(cls):
      ...     return cls(1.5)
      >>> curse(float, "myclassmethod", classmethod(myclassmethod))
      >>> float.myclassmethod()
      1.5

    Methods will be automatically bound, so don't forget to add a self
    parameter to them, like this:

      >>> def hello(self):
      ...     return self * 2
      >>> curse(str, "hello", hello)
      >>> "yo".hello()
      "yoyo"
    """
    dikt = patchable_builtin(klass)

    old_value = dikt.get(attr, None)
    old_name = "_c_%s" % attr  # do not use .format here, it breaks py2.{5,6}

    # Patch the thing
    dikt[attr] = value

    if old_value:
        dikt[old_name] = old_value

        try:
            dikt[attr].__name__ = old_value.__name__
        except (AttributeError, TypeError):  # py2.5 will raise `TypeError`
            pass
        try:
            dikt[attr].__qualname__ = old_value.__qualname__
        except AttributeError:
            pass

    ctypes.pythonapi.PyType_Modified(ctypes.py_object(klass))



from typing import Any
from typing import Dict

from ddtrace.internal.telemetry.constants import TELEMETRY_METRIC_TYPE_COUNT
from ddtrace.internal.telemetry.constants import TELEMETRY_METRIC_TYPE_DISTRIBUTIONS
from ddtrace.internal.telemetry.constants import TELEMETRY_METRIC_TYPE_GAUGE
from ddtrace.internal.telemetry.constants import TELEMETRY_METRIC_TYPE_RATE
from ddtrace.internal.telemetry.constants import TELEMETRY_NAMESPACE_TAG_APPSEC
from ddtrace.internal.telemetry.constants import TELEMETRY_NAMESPACE_TAG_TRACER
from ddtrace.internal.telemetry.constants import TELEMETRY_TYPE_DISTRIBUTION
from ddtrace.internal.telemetry.constants import TELEMETRY_TYPE_GENERATE_METRICS
from ddtrace.internal.telemetry.metrics import CountMetric
from ddtrace.internal.telemetry.metrics import DistributionMetric
from ddtrace.internal.telemetry.metrics import GaugeMetric
from ddtrace.internal.telemetry.metrics import Metric
from ddtrace.internal.telemetry.metrics import MetricTagType
from ddtrace.internal.telemetry.metrics import MetricType
from ddtrace.internal.telemetry.metrics import RateMetric


NamespaceMetricType = Dict[str, Dict[str, Dict[str, Any]]]


class TelemetryTypeError(Exception):
    pass


class MetricNamespace:
    _metrics_data = {}  # type: NamespaceMetricType
    metric_class = {
        TELEMETRY_METRIC_TYPE_COUNT: CountMetric,
        TELEMETRY_METRIC_TYPE_GAUGE: GaugeMetric,
        TELEMETRY_METRIC_TYPE_RATE: RateMetric,
        TELEMETRY_METRIC_TYPE_DISTRIBUTIONS: DistributionMetric,
    }
    metrics_type_payloads = {
        TELEMETRY_METRIC_TYPE_COUNT: TELEMETRY_TYPE_GENERATE_METRICS,
        TELEMETRY_METRIC_TYPE_GAUGE: TELEMETRY_TYPE_GENERATE_METRICS,
        TELEMETRY_METRIC_TYPE_RATE: TELEMETRY_TYPE_GENERATE_METRICS,
        TELEMETRY_METRIC_TYPE_DISTRIBUTIONS: TELEMETRY_TYPE_DISTRIBUTION,
    }

    def __init__(self):
        self._flush()

    def _flush(self):
        self._metrics_data = {
            TELEMETRY_TYPE_GENERATE_METRICS: {TELEMETRY_NAMESPACE_TAG_TRACER: {}, TELEMETRY_NAMESPACE_TAG_APPSEC: {}},
            TELEMETRY_TYPE_DISTRIBUTION: {TELEMETRY_NAMESPACE_TAG_TRACER: {}, TELEMETRY_NAMESPACE_TAG_APPSEC: {}},
        }

    def get(self):
        # type: () -> Dict[str, Any]
        namespace_metrics = self._metrics_data.copy()
        return namespace_metrics

    @staticmethod
    def validate_type_metric(metric_check, metric_to_validate):
        # type: (Metric, Metric) -> bool
        if metric_check.__class__ == metric_to_validate.__class__:
            return True
        return False

    def _add_metric(self, metric_type, namespace, name, value=1.0, tags={}, interval=10.0):
        # type: (MetricType, str,str, float, MetricTagType, float) -> None
        """
        Telemetry Metrics are stored in DD dashboards, check the metrics in datadoghq.com/metric/explorer.
        The metric will store in dashboard as "dd.instrumentation_telemetry_data." + namespace + "." + name
        """
        metric_check = self.metric_class[metric_type](  # type: ignore[abstract]
            namespace, name, tags=tags, common=True, interval=interval
        )
        metrics_type_payload = self.metrics_type_payloads[metric_type]
        # get a Metric that already exists in _metrics data or create a new one (metric_check)
        metric = self._metrics_data[metrics_type_payload][namespace].get(metric_check.id, metric_check)
        if not self.validate_type_metric(metric_check, metric):
            raise TelemetryTypeError(
                (
                    'Error: metric with name "%s" and type "%s" exists. You can\'t create a new metric '
                    'with this name an type "%s"'
                )
                % (name, metric.metric_type, metric_check.metric_type)
            )
        metric.add_point(value)
        self._metrics_data[metrics_type_payload][namespace][metric.id] = metric



import json
import logging
import os.path

import mock
import pytest
from six import ensure_binary

from ddtrace.appsec import _asm_request_context
from ddtrace.appsec._constants import DEFAULT
from ddtrace.appsec.ddwaf import DDWaf
from ddtrace.appsec.processor import AppSecSpanProcessor
from ddtrace.appsec.processor import _transform_headers
from ddtrace.constants import APPSEC_JSON
from ddtrace.constants import USER_KEEP
from ddtrace.contrib.trace_utils import set_http_meta
from ddtrace.ext import SpanTypes
from ddtrace.internal import _context
from tests.utils import override_env
from tests.utils import override_global_config
from tests.utils import snapshot


try:
    from json.decoder import JSONDecodeError
except ImportError:
    # handling python 2.X import error
    JSONDecodeError = ValueError  # type: ignore

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
RULES_GOOD_PATH = os.path.join(ROOT_DIR, "rules-good.json")
RULES_BAD_PATH = os.path.join(ROOT_DIR, "rules-bad.json")
RULES_MISSING_PATH = os.path.join(ROOT_DIR, "nonexistent")
RULES_SRB = os.path.join(ROOT_DIR, "rules-suspicious-requests.json")
RULES_SRB_RESPONSE = os.path.join(ROOT_DIR, "rules-suspicious-requests-response.json")
RULES_SRB_METHOD = os.path.join(ROOT_DIR, "rules-suspicious-requests-get.json")


@pytest.fixture
def tracer_appsec(tracer):
    with override_global_config(dict(_appsec_enabled=True)):
        yield _enable_appsec(tracer)


def _enable_appsec(tracer):
    tracer._appsec_enabled = True
    # Hack: need to pass an argument to configure so that the processors are recreated
    tracer.configure(api_version="v0.4")
    return tracer


class Config(object):
    def __init__(self):
        self.is_header_tracing_configured = False


def test_transform_headers():
    transformed = _transform_headers(
        {
            "hello": "world",
            "Foo": "bar1",
            "foo": "bar2",
            "fOO": "bar3",
            "BAR": "baz",
            "COOKIE": "secret",
        },
    )
    assert set(transformed.keys()) == {"hello", "bar", "foo"}
    assert transformed["hello"] == "world"
    assert transformed["bar"] == "baz"
    assert set(transformed["foo"]) == {"bar1", "bar2", "bar3"}


def test_enable(tracer_appsec):
    tracer = tracer_appsec

    with tracer.trace("test", span_type=SpanTypes.WEB) as span:
        set_http_meta(span, {}, raw_uri="http://example.com/.git", status_code="404")

    assert span.get_metric("_dd.appsec.enabled") == 1.0


def test_enable_custom_rules():
    with override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
        processor = AppSecSpanProcessor()

    assert processor.enabled
    assert processor.rules == RULES_GOOD_PATH


@pytest.mark.parametrize("rule,exc", [(RULES_MISSING_PATH, IOError), (RULES_BAD_PATH, ValueError)])
def test_enable_bad_rules(rule, exc, tracer):
    # with override_env(dict(DD_APPSEC_RULES=rule)):
    #     with pytest.raises(exc):
    #         _enable_appsec(tracer)

    # by default enable must not crash but display errors in the logs
    with override_global_config(dict(_raise=False)):
        with override_env(dict(DD_APPSEC_RULES=rule)):
            _enable_appsec(tracer)


def test_retain_traces(tracer_appsec):
    tracer = tracer_appsec

    with tracer.trace("test", span_type=SpanTypes.WEB) as span:
        set_http_meta(span, {}, raw_uri="http://example.com/.git", status_code="404")

    assert span.context.sampling_priority == USER_KEEP


def test_valid_json(tracer_appsec):
    tracer = tracer_appsec

    with tracer.trace("test", span_type=SpanTypes.WEB) as span:
        set_http_meta(span, {}, raw_uri="http://example.com/.git", status_code="404")

    assert "triggers" in json.loads(span.get_tag(APPSEC_JSON))


def test_header_attack(tracer_appsec):
    tracer = tracer_appsec

    with override_global_config(dict(retrieve_client_ip=True)):
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(
                span,
                Config(),
                request_headers={
                    "User-Agent": "Arachni/v1",
                    "user-agent": "aa",
                    "x-forwarded-for": "8.8.8.8",
                },
            )

        assert "triggers" in json.loads(span.get_tag(APPSEC_JSON))
        assert span.get_tag("actor.ip") == "8.8.8.8"


def test_headers_collection(tracer_appsec):
    tracer = tracer_appsec

    with tracer.trace("test", span_type=SpanTypes.WEB) as span:
        set_http_meta(
            span,
            Config(),
            raw_uri="http://example.com/.git",
            status_code="404",
            request_headers={
                "hello": "world",
                "accept": "something",
                "x-Forwarded-for": "127.0.0.1",
            },
            response_headers={
                "foo": "bar",
                "Content-Length": "500",
            },
        )

    assert span.get_tag("http.request.headers.hello") is None
    assert span.get_tag("http.request.headers.accept") == "something"
    assert span.get_tag("http.request.headers.x-forwarded-for") == "127.0.0.1"
    assert span.get_tag("http.response.headers.content-length") == "500"
    assert span.get_tag("http.response.headers.foo") is None


@snapshot(
    include_tracer=True,
    ignores=[
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
    ],
)
def test_appsec_cookies_no_collection_snapshot(tracer):
    # We use tracer instead of tracer_appsec because snapshot is looking for tracer fixture and not understands
    # other fixtures
    with override_global_config(dict(_appsec_enabled=True)):
        _enable_appsec(tracer)
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(
                span,
                {},
                raw_uri="http://example.com/.git",
                status_code="404",
                request_cookies={"cookie1": "im the cookie1"},
            )

        assert "triggers" in json.loads(span.get_tag(APPSEC_JSON))


@snapshot(
    include_tracer=True,
    ignores=[
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
    ],
)
def test_appsec_body_no_collection_snapshot(tracer):
    with override_global_config(dict(_appsec_enabled=True)):
        _enable_appsec(tracer)
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(
                span,
                {},
                raw_uri="http://example.com/.git",
                status_code="404",
                request_body={"somekey": "somekey value"},
            )

        assert "triggers" in json.loads(span.get_tag(APPSEC_JSON))


_BLOCKED_IP = "8.8.4.4"
_ALLOWED_IP = "1.1.1.1"


def test_ip_block(tracer):
    with override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)), override_global_config(dict(_appsec_enabled=True)):
        _enable_appsec(tracer)
        with _asm_request_context.asm_request_context_manager(_BLOCKED_IP, {}):
            with tracer.trace("test", span_type=SpanTypes.WEB) as span:
                set_http_meta(
                    span,
                    Config(),
                )

            assert "triggers" in json.loads(span.get_tag(APPSEC_JSON))
            assert _context.get_item("http.request.remote_ip", span) == _BLOCKED_IP
            assert _context.get_item("http.request.blocked", span)


def test_ip_not_block(tracer):
    with override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)), override_global_config(dict(_appsec_enabled=True)):
        _enable_appsec(tracer)
        with _asm_request_context.asm_request_context_manager(_ALLOWED_IP, {}):
            with tracer.trace("test", span_type=SpanTypes.WEB) as span:
                set_http_meta(
                    span,
                    Config(),
                )

            assert _context.get_item("http.request.remote_ip", span) == _ALLOWED_IP
            assert _context.get_item("http.request.blocked", span) is None


def test_ip_update_rules_and_block(tracer):
    with override_global_config(dict(_appsec_enabled=True)):
        _enable_appsec(tracer)
        tracer._appsec_processor._update_rules(
            {
                "rules_data": [
                    {
                        "data": [
                            {"value": _BLOCKED_IP},
                        ],
                        "id": "blocked_ips",
                        "type": "ip_with_expiration",
                    },
                ]
            }
        )
        with _asm_request_context.asm_request_context_manager(_BLOCKED_IP, {}):
            with tracer.trace("test", span_type=SpanTypes.WEB) as span:
                set_http_meta(
                    span,
                    Config(),
                )

                assert _context.get_item("http.request.remote_ip", span) == _BLOCKED_IP
                assert _context.get_item("http.request.blocked", span)


def test_ip_update_rules_expired_no_block(tracer):
    with override_global_config(dict(_appsec_enabled=True)):
        _enable_appsec(tracer)
        tracer._appsec_processor._update_rules(
            {
                "rules_data": [
                    {
                        "data": [
                            {"expiration": 1662804872, "value": _BLOCKED_IP},
                        ],
                        "id": "blocked_ips",
                        "type": "ip_with_expiration",
                    },
                ]
            }
        )
        with _asm_request_context.asm_request_context_manager(_BLOCKED_IP, {}):
            with tracer.trace("test", span_type=SpanTypes.WEB) as span:
                set_http_meta(
                    span,
                    Config(),
                )

            assert _context.get_item("http.request.remote_ip", span) == _BLOCKED_IP
            assert _context.get_item("http.request.blocked", span) is None


@snapshot(
    include_tracer=True,
    ignores=[
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
    ],
)
def test_appsec_span_tags_snapshot(tracer):
    with override_global_config(dict(_appsec_enabled=True)):
        _enable_appsec(tracer)
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            span.set_tag("http.url", "http://example.com/.git")
            set_http_meta(span, {}, raw_uri="http://example.com/.git", status_code="404")

        assert "triggers" in json.loads(span.get_tag(APPSEC_JSON))


@snapshot(
    include_tracer=True,
    ignores=[
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
        "meta._dd.appsec.event_rules.errors",
    ],
)
def test_appsec_span_tags_snapshot_with_errors(tracer):
    with override_global_config(dict(_appsec_enabled=True)):
        with override_env(dict(DD_APPSEC_RULES=os.path.join(ROOT_DIR, "rules-with-2-errors.json"))):
            _enable_appsec(tracer)
            with tracer.trace("test", span_type=SpanTypes.WEB) as span:
                span.set_tag("http.url", "http://example.com/.git")
                set_http_meta(span, {}, raw_uri="http://example.com/.git", status_code="404")

        assert span.get_tag(APPSEC_JSON) is None


def test_appsec_span_rate_limit(tracer):
    with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_TRACE_RATE_LIMIT="1")):
        _enable_appsec(tracer)
        with tracer.trace("test", span_type=SpanTypes.WEB) as span1:
            set_http_meta(span1, {}, raw_uri="http://example.com/.git", status_code="404")

        with tracer.trace("test", span_type=SpanTypes.WEB) as span2:
            set_http_meta(span2, {}, raw_uri="http://example.com/.git", status_code="404")
            span2.start_ns = span1.start_ns + 1

        with tracer.trace("test", span_type=SpanTypes.WEB) as span3:
            set_http_meta(span3, {}, raw_uri="http://example.com/.git", status_code="404")
            span2.start_ns = span1.start_ns + 2

        assert span1.get_tag(APPSEC_JSON) is not None
        assert span2.get_tag(APPSEC_JSON) is None
        assert span3.get_tag(APPSEC_JSON) is None


def test_ddwaf_not_raises_exception():
    with open(DEFAULT.RULES) as rules:
        rules_json = json.loads(rules.read())
        DDWaf(
            rules_json,
            ensure_binary(DEFAULT.APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP),
            ensure_binary(DEFAULT.APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP),
        )


def test_obfuscation_parameter_key_empty():
    with override_env(dict(DD_APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP="")):
        processor = AppSecSpanProcessor()

    assert processor.enabled


def test_obfuscation_parameter_value_empty():
    with override_env(dict(DD_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP="")):
        processor = AppSecSpanProcessor()

    assert processor.enabled


def test_obfuscation_parameter_key_and_value_empty():
    with override_env(
        dict(DD_APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP="", DD_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP="")
    ):
        processor = AppSecSpanProcessor()

    assert processor.enabled


def test_obfuscation_parameter_key_invalid_regex():
    with override_env(dict(DD_APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP="(")):
        processor = AppSecSpanProcessor()

    assert processor.enabled


def test_obfuscation_parameter_invalid_regex():
    with override_env(dict(DD_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP="(")):
        processor = AppSecSpanProcessor()

    assert processor.enabled


def test_obfuscation_parameter_key_and_value_invalid_regex():
    with override_env(
        dict(DD_APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP="(", DD_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP="(")
    ):
        processor = AppSecSpanProcessor()

    assert processor.enabled


def test_obfuscation_parameter_value_unconfigured_not_matching(tracer_appsec):
    tracer = tracer_appsec

    with tracer.trace("test", span_type=SpanTypes.WEB) as span:
        set_http_meta(span, Config(), raw_uri="http://example.com/.git?hello=goodbye", status_code="404")

    assert "triggers" in json.loads(span.get_tag("_dd.appsec.json"))

    assert "hello" in span.get_tag("_dd.appsec.json")
    assert "goodbye" in span.get_tag("_dd.appsec.json")
    assert "<Redacted>" not in span.get_tag("_dd.appsec.json")


def test_obfuscation_parameter_value_unconfigured_matching(tracer_appsec):
    tracer = tracer_appsec

    with tracer.trace("test", span_type=SpanTypes.WEB) as span:
        set_http_meta(span, Config(), raw_uri="http://example.com/.git?password=goodbye", status_code="404")

    assert "triggers" in json.loads(span.get_tag("_dd.appsec.json"))

    assert "password" not in span.get_tag("_dd.appsec.json")
    assert "goodbye" not in span.get_tag("_dd.appsec.json")
    assert "<Redacted>" in span.get_tag("_dd.appsec.json")


def test_obfuscation_parameter_value_configured_not_matching(tracer):
    with override_global_config(dict(_appsec_enabled=True)), override_env(
        dict(DD_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP="token")
    ):
        _enable_appsec(tracer)

        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(span, Config(), raw_uri="http://example.com/.git?password=goodbye", status_code="404")

        assert "triggers" in json.loads(span.get_tag("_dd.appsec.json"))

        assert "password" in span.get_tag("_dd.appsec.json")
        assert "goodbye" in span.get_tag("_dd.appsec.json")
        assert "<Redacted>" not in span.get_tag("_dd.appsec.json")


def test_obfuscation_parameter_value_configured_matching(tracer):
    with override_global_config(dict(_appsec_enabled=True)), override_env(
        dict(DD_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP="token")
    ):
        _enable_appsec(tracer)

        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(span, Config(), raw_uri="http://example.com/.git?token=goodbye", status_code="404")

        assert "triggers" in json.loads(span.get_tag("_dd.appsec.json"))

        assert "token" not in span.get_tag("_dd.appsec.json")
        assert "goodbye" not in span.get_tag("_dd.appsec.json")
        assert "<Redacted>" in span.get_tag("_dd.appsec.json")


def test_ddwaf_run():
    with open(RULES_GOOD_PATH) as rules:
        rules_json = json.loads(rules.read())
        _ddwaf = DDWaf(rules_json, b"", b"")
        data = {
            "server.request.query": {},
            "server.request.headers.no_cookies": {"user-agent": "werkzeug/2.1.2", "host": "localhost"},
            "server.request.cookies": {"attack": "1' or '1' = '1'"},
            "server.response.headers.no_cookies": {"content-type": "text/html; charset=utf-8", "content-length": "207"},
        }
        ctx = _ddwaf._at_request_start()
        res = _ddwaf.run(ctx, data, DEFAULT.WAF_TIMEOUT)  # res is a serialized json
        assert res.data.startswith('[{"rule":{"id":"crs-942-100"')
        assert res.runtime > 0
        assert res.total_runtime > 0
        assert res.total_runtime > res.runtime


def test_ddwaf_info():
    with open(RULES_GOOD_PATH) as rules:
        rules_json = json.loads(rules.read())
        _ddwaf = DDWaf(rules_json, b"", b"")

        info = _ddwaf.info
        assert info.loaded == 5
        assert info.failed == 0
        assert info.errors == {}
        assert info.version == "rules_good"


def test_ddwaf_info_with_2_errors():
    with open(os.path.join(ROOT_DIR, "rules-with-2-errors.json")) as rules:
        rules_json = json.loads(rules.read())
        _ddwaf = DDWaf(rules_json, b"", b"")

        info = _ddwaf.info
        assert info.loaded == 1
        assert info.failed == 2
        # Compare dict contents insensitive to ordering
        expected_dict = sorted(
            {"missing key 'conditions'": ["crs-913-110"], "missing key 'tags'": ["crs-942-100"]}.items()
        )
        assert sorted(info.errors.items()) == expected_dict
        assert info.version == "5.5.5"


def test_ddwaf_info_with_3_errors():
    with open(os.path.join(ROOT_DIR, "rules-with-3-errors.json")) as rules:
        rules_json = json.loads(rules.read())
        _ddwaf = DDWaf(rules_json, b"", b"")

        info = _ddwaf.info
        assert info.loaded == 1
        assert info.failed == 2
        assert info.errors == {"missing key 'name'": ["crs-942-100", "crs-913-120"]}


def test_ddwaf_info_with_json_decode_errors(tracer_appsec, caplog):
    tracer = tracer_appsec
    config = Config()
    config.http_tag_query_string = True

    with caplog.at_level(logging.WARNING), mock.patch(
        "ddtrace.appsec.processor.json.dumps", side_effect=JSONDecodeError("error", "error", 0)
    ), mock.patch.object(DDWaf, "info"):
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(
                span,
                config,
                method="PATCH",
                url="http://localhost/api/unstable/role_requests/dab1e9ae-9d99-11ed-bfdf-da7ad0900000?_authentication_token=2b0297348221f294de3a047e2ecf1235abb866b6",  # noqa: E501
                status_code="200",
                raw_uri="http://localhost/api/unstable/role_requests/dab1e9ae-9d99-11ed-bfdf-da7ad0900000?_authentication_token=2b0297348221f294de3a047e2ecf1235abb866b6",  # noqa: E501
                request_headers={
                    "host": "localhost",
                    "user-agent": "aa",
                    "content-length": "73",
                },
                response_headers={
                    "content-length": "501",
                    "x-ratelimit-remaining": "363",
                    "x-ratelimit-name": "role_api",
                    "x-ratelimit-limit": "500",
                    "x-ratelimit-period": "60",
                    "content-type": "application/json",
                    "x-ratelimit-reset": "16",
                },
                request_body={"_authentication_token": "2b0297348221f294de3a047e2ecf1235abb866b6"},
            )

    assert "Error parsing data ASM In-App WAF metrics report" in caplog.text


def test_ddwaf_run_contained_typeerror(tracer_appsec, caplog):
    tracer = tracer_appsec

    config = Config()
    config.http_tag_query_string = True

    with caplog.at_level(logging.DEBUG), mock.patch(
        "ddtrace.appsec.ddwaf.ddwaf_run", side_effect=TypeError("expected c_long instead of int")
    ):
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(
                span,
                config,
                method="PATCH",
                url="http://localhost/api/unstable/role_requests/dab1e9ae-9d99-11ed-bfdf-da7ad0900000?_authentication_token=2b0297348221f294de3a047e2ecf1235abb866b6",  # noqa: E501
                status_code="200",
                raw_uri="http://localhost/api/unstable/role_requests/dab1e9ae-9d99-11ed-bfdf-da7ad0900000?_authentication_token=2b0297348221f294de3a047e2ecf1235abb866b6",  # noqa: E501
                request_headers={
                    "host": "localhost",
                    "user-agent": "aa",
                    "content-length": "73",
                },
                response_headers={
                    "content-length": "501",
                    "x-ratelimit-remaining": "363",
                    "x-ratelimit-name": "role_api",
                    "x-ratelimit-limit": "500",
                    "x-ratelimit-period": "60",
                    "content-type": "application/json",
                    "x-ratelimit-reset": "16",
                },
                request_body={"_authentication_token": "2b0297348221f294de3a047e2ecf1235abb866b6"},
            )

    assert span.get_tag(APPSEC_JSON) is None
    assert "TypeError: expected c_long instead of int" in caplog.text


def test_ddwaf_run_contained_oserror(tracer_appsec, caplog):
    tracer = tracer_appsec

    config = Config()
    config.http_tag_query_string = True

    with caplog.at_level(logging.DEBUG), mock.patch(
        "ddtrace.appsec.ddwaf.ddwaf_run", side_effect=OSError("ddwaf run failed")
    ):
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(
                span,
                config,
                method="PATCH",
                url="http://localhost/api/unstable/role_requests/dab1e9ae-9d99-11ed-bfdf-da7ad0900000?_authentication_token=2b0297348221f294de3a047e2ecf1235abb866b6",  # noqa: E501
                status_code="200",
                raw_uri="http://localhost/api/unstable/role_requests/dab1e9ae-9d99-11ed-bfdf-da7ad0900000?_authentication_token=2b0297348221f294de3a047e2ecf1235abb866b6",  # noqa: E501
                request_headers={
                    "host": "localhost",
                    "user-agent": "aa",
                    "content-length": "73",
                },
                response_headers={
                    "content-length": "501",
                    "x-ratelimit-remaining": "363",
                    "x-ratelimit-name": "role_api",
                    "x-ratelimit-limit": "500",
                    "x-ratelimit-period": "60",
                    "content-type": "application/json",
                    "x-ratelimit-reset": "16",
                },
                request_body={"_authentication_token": "2b0297348221f294de3a047e2ecf1235abb866b6"},
            )

    assert span.get_tag(APPSEC_JSON) is None
    assert "OSError: ddwaf run failed" in caplog.text



from envier import En
from envier import validators


class DatabaseMonitoringConfig(En):
    __prefix__ = "dd_dbm"

    propagation_mode = En.v(
        str,
        "propagation_mode",
        default="disabled",
        help="Valid Injection Modes: disabled, service, and full",
        validator=validators.choice(["disabled", "full", "service"]),
    )


dbm_config = DatabaseMonitoringConfig()



from types import CodeType
from types import FrameType

from ddtrace.internal.logger import get_logger


log = get_logger(__name__)


cpdef _extract_class_name(frame):
    # type: (...) -> str
    """Extract class name from a frame, if possible.

    :param frame: The frame object.
    """
    if frame.f_code.co_varnames:
        argname = frame.f_code.co_varnames[0]
        try:
            value = frame.f_locals[argname]
        except KeyError:
            return ""
        try:
            if argname == "self":
                return object.__getattribute__(type(value), "__name__")  # use type() and object.__getattribute__ to avoid side-effects
            if argname == "cls":
                return object.__getattribute__(value, "__name__")
        except AttributeError:
            return ""
    return ""


cpdef traceback_to_frames(traceback, max_nframes):
    """Serialize a Python traceback object into a list of tuple of (filename, lineno, function_name).

    :param traceback: The traceback object to serialize.
    :param max_nframes: The maximum number of frames to return.
    :return: The serialized frames and the number of frames present in the original traceback.
    """
    tb = traceback
    frames = []
    nframes = 0
    while tb is not None:
        if nframes < max_nframes:
            frame = tb.tb_frame
            code = frame.f_code
            lineno = 0 if frame.f_lineno is None else frame.f_lineno
            frames.insert(0, (code.co_filename, lineno, code.co_name, _extract_class_name(frame)))
        nframes += 1
        tb = tb.tb_next
    return frames, nframes


cpdef pyframe_to_frames(frame, max_nframes):
    """Convert a Python frame to a list of frames.

    :param frame: The frame object to serialize.
    :param max_nframes: The maximum number of frames to return.
    :return: The serialized frames and the number of frames present in the original traceback."""
    # DEV: There are reports that Python 3.11 returns non-frame objects when
    # retrieving frame objects and doing stack unwinding. If we detect a
    # non-frame object we log a warning and return an empty stack, to avoid
    # reporting potentially incomplete and/or inaccurate data. This until we can
    # come to the bottom of the issue.
    if not isinstance(frame, FrameType):
        log.warning(
            "Got object of type '%s' instead of a frame object for the top frame of a thread", type(frame).__name__
        )
        return [], 0

    frames = []
    nframes = 0

    while frame is not None:
        IF PY_MAJOR_VERSION > 3 or (PY_MAJOR_VERSION == 3 and PY_MINOR_VERSION >= 11):
            if not isinstance(frame, FrameType):
                log.warning(
                    "Got object of type '%s' instead of a frame object during stack unwinding", type(frame).__name__
                )
                return [], 0
        
        if nframes < max_nframes:
            code = frame.f_code
            IF PY_MAJOR_VERSION > 3 or (PY_MAJOR_VERSION == 3 and PY_MINOR_VERSION >= 11):
                if not isinstance(code, CodeType):
                    log.warning(
                        "Got object of type '%s' instead of a code object during stack unwinding", type(code).__name__
                    )
                    return [], 0

            lineno = 0 if frame.f_lineno is None else frame.f_lineno
            frames.append((code.co_filename, lineno, code.co_name, _extract_class_name(frame)))
        nframes += 1
        frame = frame.f_back
    return frames, nframes



"""
The aws_lambda integration currently enables traces to be sent
before an impending timeout in an AWS Lambda function instrumented with the
`Datadog Lambda Python <https://github.com/DataDog/datadog-lambda-python>`_ package.

Enabling
~~~~~~~~

The aws_lambda integration is enabled automatically for AWS Lambda
functions which have been instrumented with Datadog.


Global Configuration
~~~~~~~~~~~~~~~~~~~~

This integration is configured automatically. The `datadog_lambda` package
calls ``patch_all`` when ``DD_TRACE_ENABLED`` is set to ``true``.
It's not recommended to call ``patch`` for it manually. Since it would not do
anything for other environments that do not meet the criteria above.


Configuration
~~~~~~~~~~~~~

.. important::

     You can configure some features with environment variables.

.. py:data:: DD_APM_FLUSH_DEADLINE_MILLISECONDS

   When to flush unfinished spans in an impending timeout.

   Default: AWS Lambda function timeout limit.


For additional configuration refer to
`Instrumenting Python Serverless Applications by Datadog <https://docs.datadoghq.com/serverless/installation/python>`_.
"""
from .patch import patch
from .patch import unpatch


__all__ = ["patch", "unpatch"]



VULN_INSECURE_HASHING_TYPE = "WEAK_HASH"
VULN_WEAK_CIPHER_TYPE = "WEAK_CIPHER"

EVIDENCE_ALGORITHM_TYPE = "ALGORITHM"

MD5_DEF = "md5"
SHA1_DEF = "sha1"

DES_DEF = "des"
BLOWFISH_DEF = "blowfish"
RC2_DEF = "rc2"
RC4_DEF = "rc4"
IDEA_DEF = "idea"

DEFAULT_WEAK_HASH_ALGORITHMS = {MD5_DEF, SHA1_DEF}

DEFAULT_WEAK_CIPHER_ALGORITHMS = {DES_DEF, BLOWFISH_DEF, RC2_DEF, RC4_DEF, IDEA_DEF}



# -*- coding: utf-8 -*-

import inspect

from ddtrace.debugging import safety


def test_get_args():
    def assert_args(args):
        assert set(dict(safety.get_args(inspect.currentframe().f_back)).keys()) == args

    def assert_locals(_locals):
        assert set(dict(safety.get_locals(inspect.currentframe().f_back)).keys()) == _locals

    def arg_and_kwargs(a, **kwargs):
        assert_args({"a", "kwargs"})
        assert_locals(set())

    def arg_and_args_and_kwargs(a, *ars, **kwars):
        assert_args({"a", "ars", "kwars"})
        assert_locals(set())

    def args_and_kwargs(*ars, **kwars):
        assert_args({"ars", "kwars"})
        assert_locals(set())

    def args(*ars):
        assert_args({"ars"})
        assert_locals(set())

    arg_and_kwargs(1, b=2)
    arg_and_args_and_kwargs(1, 42, b=2)
    args_and_kwargs()
    args()


# ---- Side effects ----


class SideEffects(object):
    class SideEffect(Exception):
        pass

    def __getattribute__(self, name):
        raise SideEffects.SideEffect()

    def __get__(self, instance, owner):
        raise self.SideEffect()

    @property
    def property_with_side_effect(self):
        raise self.SideEffect()


def test_get_fields_side_effects():
    assert safety.get_fields(SideEffects()) == {}


# ---- Slots ----


def test_get_fields_slots():
    class A(object):
        __slots__ = ["a"]

        def __init__(self):
            self.a = "a"

    class B(A):
        __slots__ = ["b"]

        def __init__(self):
            super(B, self).__init__()
            self.b = "b"

    assert safety.get_fields(A()) == {"a": "a"}
    assert safety.get_fields(B()) == {"a": "a", "b": "b"}



import os
import signal
import subprocess
import time
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List

import pytest
import tenacity

from ddtrace.contrib.flask.patch import flask_version
from tests.webclient import Client


DEFAULT_HEADERS = {
    "User-Agent": "python-httpx/x.xx.x",
}


@pytest.fixture
def flask_port():
    # type: () -> str
    return "8000"


@pytest.fixture
def flask_wsgi_application():
    # type: () -> str
    return "tests.contrib.flask.app:app"


@pytest.fixture
def flask_command(flask_wsgi_application, flask_port):
    # type: (str, str) -> List[str]
    cmd = "ddtrace-run flask run -h 0.0.0.0 -p %s" % (flask_port,)
    return cmd.split()


def flask_default_env(flask_wsgi_application):
    # type: (str) -> Dict[str, str]
    env = os.environ.copy()
    env.update(
        {
            # Avoid noisy database spans being output on app startup/teardown.
            "DD_TRACE_SQLITE3_ENABLED": "0",
            "FLASK_APP": flask_wsgi_application,
        }
    )
    return env


@pytest.fixture
def flask_client(flask_command, flask_port, flask_wsgi_application, flask_env_arg):
    # type: (List[str], Dict[str, str], str, Callable) -> Generator[Client, None, None]
    # Copy the env to get the correct PYTHONPATH and such
    # from the virtualenv.
    # webservers might exec or fork into another process, so we need to os.setsid() to create a process group
    # (all of which will listen to signals sent to the parent) so that we can kill the whole application.
    proc = subprocess.Popen(
        flask_command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
        env=flask_env_arg(flask_wsgi_application),
        preexec_fn=os.setsid,
        cwd=str(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))),
    )
    try:
        client = Client("http://0.0.0.0:%s" % flask_port)
        # Wait for the server to start up
        try:
            client.wait()
        except tenacity.RetryError:
            # process failed
            stdout = proc.stdout.read()
            stderr = proc.stderr.read()
            raise TimeoutError(
                "Server failed to start\n======STDOUT=====%s\n\n======STDERR=====%s\n" % (stdout, stderr)
            )
        yield client
        try:
            client.get_ignored("/shutdown")
        except Exception:
            pass
        # At this point the traces have been sent to the test agent
        # but the test agent hasn't necessarily finished processing
        # the traces (race condition) so wait just a bit for that
        # processing to complete.
        time.sleep(0.2)
    finally:
        os.killpg(proc.pid, signal.SIGKILL)
        proc.wait()


@pytest.mark.snapshot(
    ignores=["meta.flask.version"], variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)}
)
@pytest.mark.parametrize("flask_env_arg", (flask_default_env,))
def test_flask_200(flask_client):
    # type: (Client) -> None
    assert flask_client.get("/", headers=DEFAULT_HEADERS).status_code == 200


@pytest.mark.snapshot(
    ignores=["meta.flask.version"],
    variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)},
)
@pytest.mark.parametrize("flask_env_arg", (flask_default_env,))
def test_flask_stream(flask_client):
    # type: (Client) -> None
    resp = flask_client.get("/stream", headers=DEFAULT_HEADERS, stream=True)
    # read streamed reasponse, this will close the flask.response span
    assert list(resp.iter_lines()) == [b"0123456789"]
    assert resp.status_code == 200


@pytest.mark.snapshot(
    ignores=["meta.flask.version", "meta.http.useragent"],
    variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)},
)
@pytest.mark.parametrize("flask_env_arg", (flask_default_env,))
def test_flask_get_user(flask_client):
    # type: (Client) -> None
    assert flask_client.get("/identify").status_code == 200



# -*- coding: utf-8 -*-
import abc
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Text
from typing import Tuple

import six

from ddtrace.internal.telemetry.constants import TELEMETRY_METRIC_TYPE_COUNT
from ddtrace.internal.telemetry.constants import TELEMETRY_METRIC_TYPE_DISTRIBUTIONS
from ddtrace.internal.telemetry.constants import TELEMETRY_METRIC_TYPE_GAUGE
from ddtrace.internal.telemetry.constants import TELEMETRY_METRIC_TYPE_RATE


MetricType = Text
MetricTagType = Dict[str, Any]


class Metric(six.with_metaclass(abc.ABCMeta)):
    """
    Telemetry Metrics are stored in DD dashboards, check the metrics in datadoghq.com/metric/explorer
    """

    metric_type = ""
    _points = []  # type: List[Tuple[float, float]]

    def __init__(self, namespace, name, tags, common, interval=None):
        # type: (str, str, MetricTagType, bool, Optional[float]) -> None
        """
        namespace: the scope of the metric: tracer, appsec, etc.
        name: string
        tags: extra information attached to a metric
        common: set to True if a metric is common to all tracers, false if it is python specific
        interval: field set for gauge and rate metrics, any field set is ignored for count metrics (in secs)
        """
        self.name = name
        self.is_common_to_all_tracers = common
        self.interval = interval
        self.namespace = namespace
        self._tags = tags  # type: MetricTagType
        self._count = 0.0
        self._points = []  # type: List[float]

    @property
    def id(self):
        """
        https://www.datadoghq.com/blog/the-power-of-tagged-metrics/#whats-a-metric-tag
        """
        return self.name + str(self._tags)

    def __hash__(self):
        return self.id

    @abc.abstractmethod
    def add_point(self, value=1.0):
        # type: (float) -> None
        """adds timestamped data point associated with a metric"""
        pass

    def set_tags(self, tags):
        # type: (MetricTagType) -> None
        """sets a metrics tag"""
        if tags:
            for k, v in iter(tags.items()):
                self.set_tag(k, v)

    def set_tag(self, name, value):
        # type: (str, str) -> None
        """sets a metrics tag"""
        self._tags[name] = value

    def to_dict(self):
        # type: () -> Dict
        """returns a dictionary containing the metrics fields expected by the telemetry intake service"""
        data = {
            "metric": self.name,
            "type": self.metric_type,
            "common": self.is_common_to_all_tracers,
            "points": self._points,
            "tags": ["%s:%s" % (k, v) for k, v in self._tags.items()],
        }
        if self.interval is not None:
            data["interval"] = int(self.interval)
        return data


class CountMetric(Metric):
    """
    A count type adds up all the submitted values in a time interval. This would be suitable for a
    metric tracking the number of website hits, for instance.
    """

    metric_type = TELEMETRY_METRIC_TYPE_COUNT

    def __init__(self, namespace, name, tags, common, interval=None):
        super(CountMetric, self).__init__(namespace, name, tags, common, interval)
        self.interval = None

    def add_point(self, value=1.0):
        # type: (float) -> None
        """adds timestamped data point associated with a metric"""
        timestamp = time.time()
        if len(self._points) == 0:
            self._points = [[timestamp, float(value)]]  # type: ignore
        else:
            self._points[0][1] += float(value)  # type: ignore


class GaugeMetric(Metric):
    """
    A gauge type takes the last value reported during the interval. This type would make sense for tracking RAM or
    CPU usage, where taking the last value provides a representative picture of the host’s behavior during the time
    interval. In this case, using a different type such as count would probably lead to inaccurate and extreme values.
    Choosing the correct metric type ensures accurate data.
    """

    metric_type = TELEMETRY_METRIC_TYPE_GAUGE

    def add_point(self, value=1.0):
        # type: (float) -> None
        """adds timestamped data point associated with a metric"""
        timestamp = time.time()
        self._points = [(timestamp, float(value))]


class RateMetric(Metric):
    """
    The rate type takes the count and divides it by the length of the time interval. This is useful if you’re
    interested in the number of hits per second.
    """

    metric_type = TELEMETRY_METRIC_TYPE_RATE

    def add_point(self, value=1.0):
        # type: (float) -> None
        """Example:
        https://github.com/DataDog/datadogpy/blob/ee5ac16744407dcbd7a3640ee7b4456536460065/datadog/threadstats/metrics.py#L181
        """
        timestamp = time.time()
        self._count += value
        rate = (self._count / float(self.interval)) if self.interval else 0.0
        self._points = [(timestamp, rate)]


class DistributionMetric(Metric):
    """
    The rate type takes the count and divides it by the length of the time interval. This is useful if you’re
    interested in the number of hits per second.
    """

    metric_type = TELEMETRY_METRIC_TYPE_DISTRIBUTIONS

    def __init__(self, namespace, name, tags, common, interval=None):
        super(DistributionMetric, self).__init__(namespace, name, tags, common, interval)
        self.interval = None

    def add_point(self, value=1.0):
        # type: (float) -> None
        """Example:
        https://github.com/DataDog/datadogpy/blob/ee5ac16744407dcbd7a3640ee7b4456536460065/datadog/threadstats/metrics.py#L181
        """
        self._points.append(float(value))  # type: ignore

    def to_dict(self):
        # type: () -> Dict
        """returns a dictionary containing the metrics fields expected by the telemetry intake service"""
        data = {
            "metric": self.name,
            "points": self._points,
            "tags": ["%s:%s" % (k, v) for k, v in self._tags.items()],
        }
        return data



from importlib import import_module
import os
import signal

from ddtrace import tracer
from ddtrace.constants import ERROR_MSG
from ddtrace.constants import ERROR_TYPE
from ddtrace.internal.logger import get_logger
from ddtrace.internal.serverless import in_aws_lambda
from ddtrace.internal.utils import get_argument_value
from ddtrace.internal.wrapping import unwrap
from ddtrace.internal.wrapping import wrap


log = get_logger(__name__)


class TimeoutChannel:
    def __init__(self, context):
        self.crashed = False
        self.context = context

    def _handle_signal(self, sig, f):
        """
        Returns a signal of type `sig` with function `f`, if there are
        no previously defined signals.

        Else, wraps the given signal with the previously defined one,
        so no signals are overridden.
        """
        old_signal = signal.getsignal(sig)

        def wrap_signals(*args, **kwargs):
            if old_signal is not None:
                old_signal(*args, **kwargs)
            f(*args, **kwargs)

        # Return the incoming signal if any of the following cases happens:
        # - old signal does not exist,
        # - old signal is the same as the incoming, or
        # - old signal is our wrapper.
        # This avoids multiple signal calling and infinite wrapping.
        if not callable(old_signal) or old_signal == f or old_signal == wrap_signals:
            return signal.signal(sig, f)

        return signal.signal(sig, wrap_signals)

    def _start(self):
        self._handle_signal(signal.SIGALRM, self._crash_flush)

        remaining_time_in_millis = self.context.get_remaining_time_in_millis()
        apm_flush_deadline = int(os.environ.get("DD_APM_FLUSH_DEADLINE_MILLISECONDS", 100))
        apm_flush_deadline = 100 if apm_flush_deadline < 0 else apm_flush_deadline

        # TODO: Update logic to calculate an approximate of how long it will
        # take us to flush the spans on the queue.
        remaining_time_in_seconds = max(((remaining_time_in_millis - apm_flush_deadline) / 1000), 0)
        signal.setitimer(signal.ITIMER_REAL, remaining_time_in_seconds)

    def _crash_flush(self, _, __):
        """
        Tags the current root span with an Impending Timeout error.
        Finishes spans with ancestors from the current span.
        """
        self._remove_alarm_signal()
        self.crashed = True

        root_span = tracer.current_root_span()
        if root_span is not None:
            root_span.error = 1
            root_span.set_tag_str(ERROR_MSG, "Datadog detected an Impending Timeout")
            root_span.set_tag_str(ERROR_TYPE, "Impending Timeout")
        else:
            log.warning("An impending timeout was reached, but no root span was found. No error will be tagged.")

        current_span = tracer.current_span()
        if current_span is not None:
            current_span.finish_with_ancestors()

    def _remove_alarm_signal(self):
        """Removes the handler set for the signal `SIGALRM`."""
        signal.alarm(0)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)

    def stop(self):
        self._remove_alarm_signal()


class DatadogInstrumentation(object):
    """Patches an AWS Lambda handler function for Datadog instrumentation."""

    def __call__(self, func, args, kwargs):
        self.func = func
        self._before(args, kwargs)
        try:
            self.response = self.func(*args, **kwargs)
            return self.response
        except Exception:
            raise
        finally:
            self._after()

    def _before(self, args, kwargs):
        self.context = get_argument_value(args, kwargs, -1, "context")
        self.timeoutChannel = TimeoutChannel(self.context)

        self.timeoutChannel._start()

    def _after(self):
        if not self.timeoutChannel.crashed:
            self.timeoutChannel.stop()


def _modify_module_name(module_name):
    """Returns a valid modified module to get imported."""
    return ".".join(module_name.split("/"))


def _get_handler_and_module():
    """Returns the user AWS Lambda handler and module."""
    path = os.environ.get("DD_LAMBDA_HANDLER", None)
    _datadog_instrumentation = DatadogInstrumentation()

    if path is None:
        from datadog_lambda.wrapper import datadog_lambda_wrapper

        handler = getattr(datadog_lambda_wrapper, "__call__")

        return handler, datadog_lambda_wrapper, _datadog_instrumentation
    else:
        parts = path.rsplit(".", 1)
        (mod_name, handler_name) = parts
        modified_mod_name = _modify_module_name(mod_name)
        handler_module = import_module(modified_mod_name)
        handler = getattr(handler_module, handler_name)
        return handler, handler_module, _datadog_instrumentation


def _has_patch_module():
    """
    Ensures that the `aws_lambda` integration can be patched.

    It checks either the user has the DD_LAMBDA_HANDLER set correctly.
    Or if the `datadog_lambda` package is installed.
    """
    path = os.environ.get("DD_LAMBDA_HANDLER", None)
    if path is None:
        try:
            import_module("datadog_lambda.wrapper")
        except Exception:
            return False
    else:
        parts = path.rsplit(".", 1)
        if len(parts) != 2:
            return False
    return True


def patch():
    """Patches an AWS Lambda using the `datadog-lambda-py` Lambda layer."""

    # It's expected to only patch only in AWS Lambda environments.
    # The need to check if a patch module exists is to avoid patching
    # when `ddtrace` is present but not `datadog-lambda`.
    if not in_aws_lambda() and not _has_patch_module():
        return

    handler, handler_module, wrapper = _get_handler_and_module()

    if getattr(handler_module, "_datadog_patch", False):
        return
    setattr(handler_module, "_datadog_patch", True)

    wrap(handler, wrapper)


def unpatch():
    if not in_aws_lambda() and not _has_patch_module():
        return

    handler, handler_module, wrapper = _get_handler_and_module()

    if not getattr(handler_module, "_datadog_patch", False):
        return
    setattr(handler_module, "_datadog_patch", False)

    unwrap(handler, wrapper)



import logging

import pytest

from ddtrace.propagation._database_monitoring import default_sql_injector
from ddtrace.settings import _database_monitoring
from tests.utils import override_env


def test_propagation_mode_configuration():
    # Ensure Database Monitoring support is disabled by default
    # TODO: Enable DBM support by default
    assert _database_monitoring.dbm_config.propagation_mode == "disabled"

    # Ensure service is a valid injection mode
    with override_env(dict(DD_DBM_PROPAGATION_MODE="service")):
        config = _database_monitoring.DatabaseMonitoringConfig()
        assert config.propagation_mode == "service"

    # Ensure full is a valid injection mode
    with override_env(dict(DD_DBM_PROPAGATION_MODE="full")):
        config = _database_monitoring.DatabaseMonitoringConfig()
        assert config.propagation_mode == "full"

    # Ensure an invalid injection mode raises a ValueError
    with override_env(dict(DD_DBM_PROPAGATION_MODE="notaninjectionmode")):
        with pytest.raises(ValueError) as excinfo:
            _database_monitoring.DatabaseMonitoringConfig()
    assert (
        excinfo.value.args[0] == "Invalid value for environment variable DD_DBM_PROPAGATION_MODE: "
        "value must be one of ['disabled', 'full', 'service']"
    )


@pytest.mark.subprocess(env=dict(DD_DBM_PROPAGATION_MODE="disabled"))
def test_get_dbm_comment_disabled_mode():
    from ddtrace import tracer
    from ddtrace.propagation import _database_monitoring

    dbspan = tracer.trace("dbspan", service="orders-db")

    # when dbm propagation mode is disabled sqlcomments should NOT be generated
    dbm_popagator = _database_monitoring._DBM_Propagator(0, "query")
    sqlcomment = dbm_popagator._get_dbm_comment(dbspan)
    assert sqlcomment is None

    # when dbm propagation mode is disabled dbm_popagator.inject should NOT add dbm tags to args/kwargs
    args, kwargs = ("SELECT * from table;",), {}
    new_args, new_kwargs = dbm_popagator.inject(dbspan, args, kwargs)
    assert new_kwargs == kwargs
    assert new_args == args

    # when dbm propagation is disabled dbm tags should not be set
    assert dbspan.get_tag(_database_monitoring.DBM_TRACE_INJECTED_TAG) is None


@pytest.mark.subprocess(
    env=dict(
        DD_DBM_PROPAGATION_MODE="service",
        DD_SERVICE="orders-app",
        DD_ENV="staging",
        DD_VERSION="v7343437-d7ac743",
    )
)
def test_dbm_propagation_service_mode():
    from ddtrace import tracer
    from ddtrace.propagation import _database_monitoring

    dbspan = tracer.trace("dbname", service="orders-db")

    # when dbm propagation is service mode sql comments should be generated with dbm tags
    dbm_popagator = _database_monitoring._DBM_Propagator(0, "query")
    sqlcomment = dbm_popagator._get_dbm_comment(dbspan)
    assert sqlcomment == "/*dddbs='orders-db',dde='staging',ddps='orders-app',ddpv='v7343437-d7ac743'*/ "

    # when dbm propagation mode is service dbm_popagator.inject SHOULD add dbm tags to args/kwargs
    args, kwargs = ("SELECT * from table;",), {}
    new_args, new_kwargs = dbm_popagator.inject(dbspan, args, kwargs.copy())
    assert new_kwargs == {}
    assert new_args == (sqlcomment + args[0],)

    # ensure that dbm tag is not set (only required in full mode)
    assert dbspan.get_tag(_database_monitoring.DBM_TRACE_INJECTED_TAG) is None


@pytest.mark.subprocess(
    env=dict(
        DD_DBM_PROPAGATION_MODE="full",
        DD_SERVICE="orders-app",
        DD_ENV="staging",
        DD_VERSION="v7343437-d7ac743",
    )
)
def test_dbm_propagation_full_mode():
    from ddtrace import tracer
    from ddtrace.propagation import _database_monitoring

    dbspan = tracer.trace("dbname", service="orders-db")

    # when dbm propagation mode is full sql comments should be generated with dbm tags and traceparent keys
    dbm_popagator = _database_monitoring._DBM_Propagator(0, "procedure")
    sqlcomment = dbm_popagator._get_dbm_comment(dbspan)
    # assert tags sqlcomment contains the correct value
    assert (
        sqlcomment
        == "/*dddbs='orders-db',dde='staging',ddps='orders-app',ddpv='v7343437-d7ac743',traceparent='%s'*/ "
        % (dbspan.context._traceparent,)
    )

    # when dbm propagation mode is full dbm_popagator.inject SHOULD add dbm tags and traceparent to args/kwargs
    args, kwargs = tuple(), {"procedure": "SELECT * from table;"}
    new_args, new_kwargs = dbm_popagator.inject(dbspan, args, kwargs.copy())
    # assert that keyword was updated
    assert new_kwargs == {"procedure": sqlcomment + kwargs["procedure"]}
    # assert that args remain unchanged
    assert new_args == args

    # ensure that dbm tag is set (required for full mode)
    assert dbspan.get_tag(_database_monitoring.DBM_TRACE_INJECTED_TAG) == "true"


def test_default_sql_injector(caplog):
    # test sql injection with unicode str
    dbm_comment = "/*dddbs='orders-db'*/ "
    str_query = "select * from table;"
    assert default_sql_injector(dbm_comment, str_query) == "/*dddbs='orders-db'*/ select * from table;"

    # test sql injection with uft-8 byte str query
    dbm_comment = "/*dddbs='orders-db'*/ "
    str_query = "select * from table;".encode("utf-8")
    assert default_sql_injector(dbm_comment, str_query) == b"/*dddbs='orders-db'*/ select * from table;"

    # test sql injection with a non supported type
    with caplog.at_level(logging.INFO):
        dbm_comment = "/*dddbs='orders-db'*/ "
        non_string_object = object()
        result = default_sql_injector(dbm_comment, non_string_object)
        assert result == non_string_object

    assert "Linking Database Monitoring profiles to spans is not supported for the following query type:" in caplog.text



from typing import Dict

from ddtrace.internal.compat import ensure_pep562
from ddtrace.vendor.debtcollector import deprecate


# tags
QUERY = "sql.query"  # the query text
ROWS = "sql.rows"  # number of rows returned by a query
DB = "sql.db"  # the name of the database


def normalize_vendor(vendor):
    # type: (str) -> str
    """Return a canonical name for a type of database."""
    if not vendor:
        return "db"  # should this ever happen?
    elif "sqlite" in vendor:
        return "sqlite"
    elif "postgres" in vendor or vendor == "psycopg2":
        return "postgres"
    else:
        return vendor


def parse_pg_dsn(dsn):
    # type: (str) -> Dict[str, str]
    """
    Return a dictionary of the components of a postgres DSN.
    >>> parse_pg_dsn('user=dog port=1543 dbname=dogdata')
    {'user':'dog', 'port':'1543', 'dbname':'dogdata'}
    """
    return dict(_.split("=", 1) for _ in dsn.split())


def __getattr__(name):
    if name == "ROWS":
        deprecate(
            ("%s.%s is deprecated" % (__name__, name)),
            postfix=". Use ddtrace.ext.db.ROWCOUNT instead.",
            removal_version="2.0.0",
        )
        return "sql.rows"

    if name in globals():
        return globals()[name]

    raise AttributeError("'%s' has no attribute '%s'", __name__, name)


ensure_pep562(__name__)



#!/usr/bin/env python3
# -*- encoding: utf-8 -*-
import sys

import pytest

from ddtrace.appsec.iast._input_info import Input_info


@pytest.mark.parametrize(
    "obj, kwargs",
    [
        (3.5, {}),
        ("Hi", {}),
        ("🙀", {}),
        (b"Hi", {}),
        (b"Hi", {"encoding": "utf-8", "errors": "strict"}),
        (b"Hi", {"encoding": "utf-8", "errors": "ignore"}),
        ({"a": "b", "c": "d"}, {}),
        ({"a", "b", "c", "d"}, {}),
        (("a", "b", "c", "d"), {}),
        (["a", "b", "c", "d"], {}),
    ],
)
@pytest.mark.skipif(sys.version_info < (3, 6, 0), reason="Python 3.6+ only")
def test_str_aspect(obj, kwargs):
    import ddtrace.appsec.iast._ast.aspects as ddtrace_aspects

    assert ddtrace_aspects.str_aspect(obj, **kwargs) == str(obj, **kwargs)


@pytest.mark.parametrize(
    "obj, kwargs, should_be_tainted",
    [
        (3.5, {}, False),
        (u"Hi", {}, True),
        ("🙀", {}, True),
        (b"Hi", {}, True),
        (bytearray(b"Hi"), {}, True),
        (b"Hi", {"encoding": "utf-8", "errors": "strict"}, True),
        (b"Hi", {"encoding": "utf-8", "errors": "ignore"}, True),
        ({"a": "b", "c": "d"}, {}, False),
        ({"a", "b", "c", "d"}, {}, False),
        (("a", "b", "c", "d"), {}, False),
        (["a", "b", "c", "d"], {}, False),
    ],
)
@pytest.mark.skipif(sys.version_info < (3, 6, 0), reason="Python 3.6+ only")
def test_str_aspect_tainting(obj, kwargs, should_be_tainted):
    import ddtrace.appsec.iast._ast.aspects as ddtrace_aspects
    from ddtrace.appsec.iast._taint_tracking import clear_taint_mapping
    from ddtrace.appsec.iast._taint_tracking import is_pyobject_tainted
    from ddtrace.appsec.iast._taint_tracking import setup
    from ddtrace.appsec.iast._taint_tracking import taint_pyobject

    setup(bytes.join, bytearray.join)
    clear_taint_mapping()
    if should_be_tainted:
        obj = taint_pyobject(obj, Input_info("test_str_aspect_tainting", obj, 0))

    result = ddtrace_aspects.str_aspect(obj, **kwargs)
    assert is_pyobject_tainted(result) == should_be_tainted

    assert result == str(obj, **kwargs)



import inspect
import threading
from uuid import uuid4

import attr
import mock

from ddtrace.debugging._capture.collector import CapturedEventCollector
from ddtrace.debugging._capture.model import CaptureState
from ddtrace.debugging._capture.snapshot import Snapshot
from ddtrace.debugging._probe.model import DDExpression
from tests.debugging.utils import create_snapshot_line_probe


class MockLimiter:
    def limit(self):
        return


@attr.s
class MockFrame(object):
    f_locals = attr.ib(type=dict)


@attr.s
class MockThread(object):
    ident = attr.ib(type=int)
    name = attr.ib(type=str)


def mock_encoder(wraps=None):
    encoder = mock.Mock(wraps=wraps)
    snapshot_encoder = mock.Mock()
    encoder._encoders = {Snapshot: snapshot_encoder}

    return encoder, snapshot_encoder


def test_collector_cond():
    encoder, _ = mock_encoder()

    collector = CapturedEventCollector(encoder=encoder)

    snapshot1 = Snapshot(
        probe=create_snapshot_line_probe(
            probe_id=uuid4(),
            source_file="file.py",
            line=123,
            condition=DDExpression("a not null", lambda _: _["a"] is not None),
        ),
        frame=MockFrame(dict(a=42)),
        args=[("a", 42)],
        thread=MockThread(-1, "MainThread"),
    )
    snapshot1.line({"a": 42, "c": True})
    collector.push(snapshot1)

    snapshot2 = Snapshot(
        probe=create_snapshot_line_probe(
            probe_id=uuid4(),
            source_file="file.py",
            line=123,
            condition=DDExpression("b not null", lambda _: _["b"] is not None),
        ),
        frame=MockFrame(dict(b=None)),
        args=[("b", None)],
        thread=MockThread(-2, "WorkerThread"),
    )
    snapshot2.line({"b": None, "c": True})
    collector.push(snapshot2)

    encoder.put.assert_called_once()


def test_collector_collect_enqueue_only_commit_state():
    encoder, _ = mock_encoder()

    collector = CapturedEventCollector(encoder=encoder)
    for i in range(10):
        mockedEvent = mock.Mock()
        with collector.attach(mockedEvent):
            mockedEvent.enter.assert_called_once()
            mockedEvent.state = CaptureState.DONE_AND_COMMIT if i % 2 == 0 else CaptureState.SKIP_COND
        mockedEvent.exit.assert_called_once()

    assert len(encoder.put.mock_calls) == 5


def test_collector_push_enqueue():
    encoder, _ = mock_encoder()

    collector = CapturedEventCollector(encoder=encoder)
    for _ in range(10):
        snapshot = Snapshot(
            probe=create_snapshot_line_probe(probe_id=uuid4(), source_file="file.py", line=123),
            frame=inspect.currentframe(),
            thread=threading.current_thread(),
        )
        snapshot.line()
        collector.push(snapshot)

    assert len(encoder.put.mock_calls) == 10
